"""
This module preprocesses Sentinel-1 data. Using SNAP GPT.

TODO:
    - Enable support for processing only VV or VH.
    - Improve documentation
"""

from sys import platform
import os
import shutil
from glob import glob
from multiprocessing import cpu_count
from concurrent.futures import ThreadPoolExecutor
from typing import List, Union, Optional

import numpy as np

from buteo.raster import (
    raster_to_array,
    array_to_raster,
    raster_to_metadata,
)
from buteo.vector.metadata import vector_to_metadata
from buteo.utils.utils_gdal import _check_is_raster, _check_is_vector


def _find_gpt(
    test_gpt_path: str,
) -> str:
    """ Internal. """
    gpt = os.path.realpath(os.path.abspath(os.path.expanduser(test_gpt_path)))
    if not os.path.exists(gpt):
        possible_locations = [
            "~/esa_snap/bin/gpt",
            "~/snap/bin/gpt",
            "/opt/esa_snap/bin/gpt",
            "/opt/snap/bin/gpt",
            "C:/Program Files/snap/bin/gpt.exe",
            '"C:/Program Files/snap/bin/gpt.exe"',
        ]

        for loc in possible_locations:
            gpt = os.path.realpath(os.path.abspath(os.path.expanduser(loc)))
            if os.path.exists(gpt):
                return gpt

        assert os.path.exists(gpt), "Graph processing tool not found."

    else:
        return gpt


def _backscatter_step1(
    zip_file: str,
    out_path: str,
    gpt_path: str ="~/snap/bin/gpt",
    extent: Optional[List[Union[int, float]]]=None,
    tmp_folder=None,
) -> str:
    """ Internal. """
    graph = "backscatter_step1.xml"

    # Get absolute location of graph processing tool
    gpt = _find_gpt(gpt_path)

    out_path_ext = out_path + ".dim"
    if os.path.exists(out_path_ext):
        print(f"{out_path_ext} already processed")
        return out_path_ext

    xmlfile = os.path.join(os.path.dirname(__file__), f"./graphs/{graph}")

    snap_graph_step1 = open(xmlfile, "r")
    snap_graph_step1_str = snap_graph_step1.read()
    snap_graph_step1.close()

    if extent is not None:
        if _check_is_vector(extent):
            metadata = vector_to_metadata(extent)
        elif _check_is_raster(extent):
            metadata = raster_to_metadata(extent)
        elif isinstance(extent, str):
            metadata = raster_to_metadata(extent)
        else:
            raise ValueError("Extent must be a vector, raster or a path to a raster.")

        interest_area = metadata["extent_wkt_latlng"]

    else:
        interest_area = "POLYGON ((-180.0 -90.0, 180.0 -90.0, 180.0 90.0, -180.0 90.0, -180.0 -90.0))"

    snap_graph_step1_str = snap_graph_step1_str.replace("${extent}", interest_area)
    snap_graph_step1_str = snap_graph_step1_str.replace("${inputfile}", zip_file)
    snap_graph_step1_str = snap_graph_step1_str.replace("${outputfile}", out_path)

    xmlfile = tmp_folder + os.path.basename(out_path) + "_graph.xml"

    f = open(xmlfile, "w")
    f.write(snap_graph_step1_str)
    f.close()

    command = [
        gpt,
        os.path.abspath(xmlfile),
        f"-q {cpu_count()}",
    ]

    if platform == "linux" or platform == "linux2":
        cmd = " ".join(command)
    else:
        cmd = f'cmd /c {" ".join(command)}'

    os.system(cmd)

    return out_path_ext

def _backscatter_step2(
    dim_file: str,
    out_path: str,
    speckle_filter: bool = False,
    epsg: Optional[int] = None,
    gpt_path: str = "~/snap/bin/gpt",
    tmp_folder: Optional[str] = None,
) -> str:
    """ Internal. """
    graph = "backscatter_step2.xml"
    if speckle_filter:
        graph = "backscatter_step2_speckle.xml"

    # Get absolute location of graph processing tool
    gpt = _find_gpt(gpt_path)

    out_path_ext = out_path + ".dim"
    if os.path.exists(out_path_ext):
        print(f"{out_path_ext} already processed")
        return out_path_ext

    xmlfile = os.path.join(os.path.dirname(__file__), f"./graphs/{graph}")

    snap_graph_step2 = open(xmlfile, "r")
    snap_graph_step2_str = snap_graph_step2.read()
    snap_graph_step2.close()

    if epsg is None:
        use_epsg = "AUTO:42001"
    else:
        use_epsg = f"EPSG:{epsg}"

    snap_graph_step2_str = snap_graph_step2_str.replace("${epsg}", use_epsg)
    snap_graph_step2_str = snap_graph_step2_str.replace("${inputfile}", dim_file)
    snap_graph_step2_str = snap_graph_step2_str.replace("${outputfile}", out_path)

    xmlfile = tmp_folder + os.path.basename(out_path) + "_graph.xml"

    f = open(xmlfile, "w")
    f.write(snap_graph_step2_str)
    f.close()

    command = [
        gpt,
        os.path.abspath(xmlfile),
        f"-q {cpu_count()}",
    ]

    if platform == "linux" or platform == "linux2":
        cmd = " ".join(command)
    else:
        cmd = f'cmd /c {" ".join(command)}'

    snap_graph_step2 = open(xmlfile, "r")
    snap_graph_step2_str = snap_graph_step2.read()
    snap_graph_step2.close()

    xmlfile = tmp_folder + os.path.basename(out_path) + "_graph.xml"

    f = open(xmlfile, "w")
    f.write(snap_graph_step2_str)
    f.close()

    os.system(cmd)

    return out_path_ext


def _convert_to_tiff(
    dim_file: str,
    out_folder: str,
    decibel: bool = False,
    use_nodata: bool = True,
    nodata_value: Union[int, float] = -9999.0,
) -> List[str]:
    """ Internal. """
    data_folder = os.path.splitext(dim_file)[0] + ".data/"
    name = os.path.splitext(os.path.basename(dim_file))[0].replace("_step_2", "")

    vh_path = data_folder + "Gamma0_VH.img"
    vv_path = data_folder + "Gamma0_VV.img"

    out_paths = [
        out_folder + name + "_Gamma0_VH.tif",
        out_folder + name + "_Gamma0_VV.tif",
    ]

    if os.path.exists(out_paths[0]) and os.path.exists(out_paths[1]):
        print(f"{name} already processed")
        return out_paths

    vh = raster_to_array(vh_path)
    vv = raster_to_array(vv_path)

    if use_nodata:
        vh = np.ma.masked_equal(vh, 0.0, copy=False)
        vv = np.ma.masked_equal(vv, 0.0, copy=False)

        vh = np.nan_to_num(vh)
        vv = np.nan_to_num(vv)

        vh = np.ma.masked_equal(vh.filled(nodata_value), nodata_value)
        vv = np.ma.masked_equal(vv.filled(nodata_value), nodata_value)

    if decibel:
        with np.errstate(divide="ignore", invalid="ignore"):
            if use_nodata:

                vh = np.ma.multiply(np.ma.log10(np.ma.abs(vh)), 10)
                vv = np.ma.multiply(np.ma.log10(np.ma.abs(vv)), 10)
            else:
                vh = np.multiply(np.log10(np.abs(vh)), 10)
                vv = np.multiply(np.log10(np.abs(vv)), 10)

    array_to_raster(vh, reference=vh_path, out_path=out_paths[0])
    array_to_raster(vv, reference=vv_path, out_path=out_paths[1])

    return out_paths


def _clear_tmp_folder(
    tmp_folder: str,
) -> None:
    """ Internal. """
    try:
        tmp_files = glob(tmp_folder + "*.dim")
        for f in tmp_files:
            os.remove(f)
        tmp_files = glob(tmp_folder + "*.data")
        for f in tmp_files:
            shutil.rmtree(f)
    except Exception:
        print("Error while cleaning tmp files.")


def s1_backscatter(
    zip_file: str,
    out_path: str,
    tmp_folder: str,
    extent: Optional[List[Union[int, float]]] = None,
    epsg: Optional[int] = None,
    use_nodata: bool = True,
    nodata_value: Union[int, float] = -9999.0,
    speckle_filter: bool = False,
    decibel: bool = False,
    clean_tmp: bool = False,
    gpt_path: str = "~/snap/bin/gpt",
):
    """
    Calculate backscatter from Sentinel-1 GRD data.
    
    Parameters
    ----------
    zip_file : str
        Path to the Sentinel-1 zip file.

    out_path : str
        Path to the output folder.

    tmp_folder : str
        Path to the temporary folder.

    extent : list, optional
        The extent of the output raster in the format [xmin, xmax, ymin, ymax].
        Default: None

    epsg : int, optional
        The EPSG code of the output raster. Default: None

    use_nodata : bool, optional
        If True, the output raster will have a nodata value. Default: True

    nodata_value : int or float, optional
        The nodata value of the output raster. Default: -9999.0

    speckle_filter : bool, optional
        If True, a speckle filter will be applied. Default: False

    decibel : bool, optional
        If True, the output raster will be in decibel. Default: False

    clean_tmp : bool, optional
        If True, the temporary files will be deleted after processing. Default: False

    gpt_path : str, optional
        Path to the GPT executable. Default: "~/snap/bin/gpt"

    Returns
    -------
    list[str]
        A list with the paths to the output rasters.
    """
    base = os.path.splitext(os.path.splitext(os.path.basename(zip_file))[0])[0]
    name1 = base + "_step_1"

    vh = out_path + base + "_Gamma0_VH.tif"
    vv = out_path + base + "_Gamma0_VV.tif"

    if os.path.exists(vh) and os.path.exists(vv):
        if clean_tmp:
            _clear_tmp_folder(tmp_folder)
        print(f"{zip_file} already processed")
        return [vh, vv]

    def step1_helper(args=[]):
        return _backscatter_step1(
            args[0], args[1], gpt_path=args[2], extent=args[3], tmp_folder=args[4]
        )

    try:
        p = ThreadPoolExecutor(1)

        step1 = p.submit(
            step1_helper,
            args=[zip_file, tmp_folder + name1, gpt_path, extent, tmp_folder],
        )
        step1 = step1.result(timeout=60 * 10)

    except Exception as e:
        if clean_tmp:
            _clear_tmp_folder(tmp_folder)

        raise Exception(f"{zip_file} failed to complete step 1., {e}")

    name2 = base + "_step_2"

    def step2_helper(args=[]):
        return _backscatter_step2(
            args[0],
            args[1],
            speckle_filter=args[2],
            epsg=epsg,
            gpt_path=args[3],
            tmp_folder=args[4],
        )

    try:
        p = ThreadPoolExecutor(1)

        step2 = p.submit(
            step2_helper,
            args=(step1, tmp_folder + name2, speckle_filter, gpt_path, tmp_folder),
        )
        step2 = step2.result(timeout=60 * 60)

    except Exception as e:
        if clean_tmp:
            _clear_tmp_folder(tmp_folder)

        raise Exception(f"{zip_file} failed to complete step 2., {e}")

    print("Generating .tif files from step2.")
    converted = _convert_to_tiff(
        step2, out_path, decibel, use_nodata=use_nodata, nodata_value=nodata_value
    )

    if clean_tmp:
        _clear_tmp_folder(tmp_folder)

    return converted



"""
### Perform convolutions on arrays.  ###
"""

# Standard Library
import sys; sys.path.append("../../")
from typing import Union

# External
import numpy as np
from numba import jit, prange

# Internal
from buteo.utils.utils_base import _type_check
from buteo.array.convolution_funcs import _hood_to_value



def pad_array(
    arr: np.ndarray,
    pad_size: int = 1,
    method: str = "same",
    constant_value: Union[float, int] = 0.0,
) -> np.ndarray:
    """
    Create a padded view of an array using SAME padding.

    Parameters
    ----------
    arr : numpy.ndarray
        The input array to be padded.

    pad_size : int, optional
        The number of padding elements to add to each side of the array.
        Default: 1.

    method : str, optional
        The padding method to use. Default: "same". Other options are
        "edge" and "constant".

    constant_value : int, optional
        The constant value to use when padding with "constant". Default: 0.

    Returns
    -------
    numpy.ndarray
        A padded view of the input array.

    Notes
    -----
    This function creates a padded view of an array using SAME padding, which
    adds padding elements to each side of the array so that the output shape
    is the same as the input shape. The amount of padding is determined by the
    `pad_size` parameter. The padding method can be one of three options: "same"
    (the default), "edge", or "constant". If "constant" padding is used, the
    `constant_value` parameter specifies the value to use.
    """
    _type_check(arr, [np.ndarray], "arr")
    _type_check(pad_size, [int], "pad_size")
    _type_check(method, [str], "method")

    assert pad_size >= 0, "pad_size must be a non-negative integer"
    method = method.lower()
    assert method in ["same", "edge", "constant"], "method must be one of ['same', 'edge', 'constant']"

    if method in ["same", "edge"]:
        padded_view = np.pad(
            arr,
            pad_width=((pad_size, pad_size), (pad_size, pad_size), (0, 0)),
            mode='edge',
        )
    elif method in "constant":
        padded_view = np.pad(
            arr,
            pad_width=((pad_size, pad_size), (pad_size, pad_size), (0, 0)),
            mode='constant',
            constant_values=constant_value,
        )

    return padded_view


@jit(nopython=True, parallel=True, nogil=False, fastmath=True, cache=True)
def _convolve_array_2D(
    arr: np.ndarray,
    offsets: np.ndarray,
    weights: np.ndarray,
    method: int = 1,
    nodata: bool = False,
    nodata_value: float = -9999.9,
    func_value: Union[int, float] = 0.5,
) -> np.ndarray:
    """
    Internal function for convolving a 2D array.
    Input should be float32.
    """
    result = np.zeros((arr.shape[0], arr.shape[1]), dtype="float32")
    hood_size = len(offsets)
    weights_total = np.sum(weights)

    nodata_value = np.float32(nodata_value)

    for idx_y in prange(arr.shape[0]):
        for idx_x in range(arr.shape[1]):
            hood_values = np.zeros(hood_size, dtype="float32")
            hood_weights = np.zeros(hood_size, dtype="float32")
            hood_normalise = False
            hood_center = 0
            hood_count = 0

            if nodata and arr[idx_y, idx_x] == nodata_value:
                result[idx_y, idx_x] = nodata_value
                continue

            for idx_h in range(0, hood_size):
                hood_y = idx_y + offsets[idx_h][0]
                hood_x = idx_x + offsets[idx_h][1]

                if hood_y < 0 or hood_y >= arr.shape[0]:
                    hood_normalise = True
                    continue

                if hood_x < 0 or hood_x >= arr.shape[1]:
                    hood_normalise = True
                    continue

                if nodata and arr[hood_y, hood_x] == nodata_value:
                    hood_normalise = True
                    continue

                hood_values[hood_count] = arr[hood_y, hood_x]
                hood_weights[hood_count] = weights[idx_h]
                hood_count += 1

                if offsets[idx_h][0] == 0 and offsets[idx_h][1] == 0:
                    hood_center = hood_count - 1

            if hood_count == 0:
                result[idx_y, idx_x] = nodata_value
                continue

            hood_values = hood_values[:hood_count]
            hood_weights = hood_weights[:hood_count]

            if hood_normalise:
                hood_weights /= np.sum(hood_weights)
                hood_weights *= weights_total

            result[idx_y, idx_x] = _hood_to_value(
                method,
                hood_values,
                hood_weights,
                nodata_value,
                hood_center,
                func_value,
            )

    return result


@jit(nopython=True, nogil=True, cache=True, fastmath=True, parallel=True)
def convolve_array_simple(
    arr: np.ndarray,
    offsets: np.ndarray,
    weights: np.ndarray,
) -> np.ndarray:
    """
    Convolve a kernel with an array using a simple method.
    Array must be 2D (height, width).
    The function ignores nodata values and pads the array with 'same'.
    It only supports 'summation' method.

    Parameters
    ----------
    arr : numpy.ndarray
        The array to convolve.

    offsets : numpy.ndarray
        The offsets of the kernel.

    weights : numpy.ndarray
        The weights of the kernel.

    Returns
    -------
    numpy.ndarray
        The convolved array.
    """
    result = np.zeros(arr.shape, dtype=np.float32)

    for col in prange(arr.shape[0]):
        for row in range(arr.shape[1]):

            result_value = 0.0
            for i in range(offsets.shape[0]):
                new_col = col + offsets[i, 0]
                new_row = row + offsets[i, 1]

                if new_col < 0:
                    new_col = 0
                elif new_col >= arr.shape[0]:
                    new_col = arr.shape[0] - 1

                if new_row < 0:
                    new_row = 0
                elif new_row >= arr.shape[1]:
                    new_row = arr.shape[1] - 1

                result_value += arr[new_col, new_row] * weights[i]

            result[col, row] = result_value

    return result


@jit(nopython=True, parallel=True, nogil=False, fastmath=True, cache=True)
def _convolve_array_channels_HWC(
    arr: np.ndarray,
    method: int = 1,
    nodata: bool = False,
    nodata_value: float = -9999.9,
    func_value: Union[int, float] = 0.5,
) -> np.ndarray:
    """
    Internal function for convoling a 3D array along its channels.
    Input should be float32. Channel-last version.
    """
    result = np.zeros((arr.shape[0], arr.shape[1], 1), dtype="float32")

    hood_size = arr.shape[2]
    center_idx = int(hood_size / 2)

    for idx_y in prange(arr.shape[0]):
        for idx_x in range(arr.shape[1]):
            hood_values = np.zeros(hood_size, dtype="float32")
            hood_weights = np.ones(hood_size, dtype="float32")
            hood_count = 0

            for idx_c in range(hood_size):
                if nodata and arr[idx_y, idx_x, idx_c] == nodata_value:
                    continue

                hood_values[hood_count] = arr[idx_y, idx_x, idx_c]
                hood_count += 1

            if hood_count == 0:
                result[idx_y, idx_x, 0] = nodata_value
                continue

            hood_values = hood_values[:hood_count]
            hood_weights = hood_weights[:hood_count] / hood_weights[:hood_count].sum()

            result[idx_y, idx_x, 0] = _hood_to_value(
                method,
                hood_values,
                hood_weights,
                nodata_value,
                center_idx,
                func_value,
            )

    return result


@jit(nopython=True, parallel=True, nogil=False, fastmath=True, cache=True)
def _convolve_array_channels_CHW(
    arr: np.ndarray,
    method: int = 1,
    nodata: bool = False,
    nodata_value: float = -9999.9,
    func_value: Union[int, float] = 0.5,
) -> np.ndarray:
    """
    Internal function for convoling a 3D array along its channels.
    Input should be float32. Channel-first version.
    """
    result = np.zeros((1, arr.shape[0], arr.shape[1]), dtype="float32")

    hood_size = arr.shape[0]
    center_idx = int(hood_size / 2)

    for idx_y in prange(arr.shape[1]):
        for idx_x in range(arr.shape[2]):
            hood_count = 0
            hood_values = np.zeros(hood_size, dtype="float32")
            hood_weights = np.ones(hood_size, dtype="float32")

            for idx_c in range(hood_size):
                if nodata and arr[idx_c, idx_y, idx_x] == nodata_value:
                    continue
                hood_values[hood_count] = arr[idx_c, idx_y, idx_x]
                hood_count += 1

            if hood_count == 0:
                result[0, idx_y, idx_x] = nodata_value
                continue

            hood_values = hood_values[:hood_count]
            hood_weights = hood_weights[:hood_count] / hood_weights[:hood_count].sum()

            result[0, idx_y, idx_x] = _hood_to_value(
                method,
                hood_values,
                hood_weights,
                nodata_value,
                center_idx,
                func_value,
            )

    return result


def convolve_array_channels(
    arr: np.ndarray,
    method: int = 1,
    nodata: bool = False,
    nodata_value: float = -9999.9,
    func_value: Union[int, float] = 0.5,
    channel_last: bool = True,
) -> np.ndarray:
    """
    Convolves a 3D array along its channels.
    Useful for 'collapsing' a 3D array into a 2D array.

    Parameters
    ----------
    arr : np.ndarray
        A 3D array.

    method : int, optional
        The method to use for convolving the array.

        The following methods are valid:
        1. sum
        2. max
        3. min
        4. mean
        5. median
        6. variance
        7. standard deviation
        8. contrast
        9. mode
        10. median absolute deviation (mad)
        11. z-score
        12. z-score (mad)
        13. sigma lee
        14. quantile
        15. occurances
        16. feather
        17. roughness
        18. roughness tri
        19. roughness tpi

        Default: 1.
    
    nodata : bool, optional
        Whether to use nodata values in the convolution. Default: False.
    
    nodata_value : float, optional
        The nodata value to use in the convolution. Default: -9999.9.
    
    func_value : int or float, optional
        The value to use in the convolution. Default: 0.5.
    
    channel_last : bool, optional
        Whether the channels are the last axis in the array. Default: True.

    Returns
    -------
    np.ndarray
        The convolved array.
    """
    _type_check(arr, [np.ndarray], "arr")
    _type_check(method, [int], "method")
    _type_check(nodata, [bool], "nodata")
    _type_check(nodata_value, [float], "nodata_value")
    _type_check(func_value, [int, float], "value")

    assert arr.ndim == 3, "arr must be a 3D array"
    assert method in range(1, 18), "method must be between 1 and 17"

    if arr.shape[2] == 1:
        return arr

    arr = arr.astype(np.float32, copy=False)

    if channel_last:
        return _convolve_array_channels_HWC(
            arr,
            method=method,
            nodata=nodata,
            nodata_value=nodata_value,
            func_value=func_value,
        )
    else:
        return _convolve_array_channels_CHW(
            arr,
            method=method,
            nodata=nodata,
            nodata_value=nodata_value,
            func_value=func_value,
        )


def convolve_array(
    arr: np.ndarray,
    offsets: np.ndarray,
    weights: np.ndarray,
    method: int = 1,
    nodata: bool = False,
    nodata_value: float = -9999.9,
    func_value: Union[int, float] = 0.5,
    channel_last: bool = True,
) -> np.ndarray:
    """
    Convolve an array using a set of offsets and weights.
    Array can be 2D or 3D.

    Parameters
    ----------
    arr : numpy.ndarray
        The input array to be convolved.

    offsets : list of tuples
        The list of pixel offsets to use in the convolution. Each tuple should be in the
        format (row_offset, col_offset, depth_offset), where row_offset and col_offset
        are the row and column offsets from the center pixel, and depth_offset is the
        depth offset if the input array has more than two dimensions.

    weights : list of floats
        The list of weights to use in the convolution. The length of the weights list should
        be the same as the length of the offsets list.

    method : int, optional
        The convolution method to use. Default: 1.
        The following methods are valid:
        1. sum
        2. max
        3. min
        4. mean
        5. median
        6. variance
        7. standard deviation
        8. contrast
        9. mode
        10. median absolute deviation (mad)
        11. z-score
        12. z-score (mad)
        13. sigma lee
        14. quantile
        15. occurances
        16. feather
        17. roughness
        18. roughness tri
        19. roughness tpi

    nodata : bool, optional
        If True, treat the nodata value as a valid value. Default: False.

    nodata_value : float, optional
        The nodata value to use when computing the result. Default: -9999.9.

    func_value : int or float, optional
        The value to use for pixels where the kernel extends outside the input array.
        If None, use the edge value. Default: 0.5.
    
    channel_last : bool, optional
        Whether the channels are the last axis in the array. Default: True.

    Returns
    -------
    numpy.ndarray
        The convolved array.

    Notes
    -----
    This function convolves an array using a set of offsets and weights. The function supports
    different convolution methods, including nearest, linear, and cubic. The function can also
    handle nodata values and cases where the kernel extends outside the input array.
    """
    _type_check(arr, [np.ndarray], "arr")
    _type_check(offsets, [np.ndarray], "offsets")
    _type_check(weights, [np.ndarray], "weights")
    _type_check(method, [int], "method")
    _type_check(nodata, [bool], "nodata")
    _type_check(nodata_value, [float], "nodata_value")
    _type_check(func_value, [int, float, type(None)], "value")
    _type_check(channel_last, [bool], "channel_last")

    assert len(offsets) == len(weights), "offsets and weights must be the same length"
    assert arr.ndim in [2, 3], "arr must be 2 or 3 dimensional"
    assert method in range(1, 18), "method must be between 1 and 17"

    arr = arr.astype(np.float32, copy=False)

    if arr.ndim == 2:
        return _convolve_array_2D(
            arr,
            offsets,
            weights,
            method=method,
            nodata=nodata,
            nodata_value=nodata_value,
            func_value=func_value,
        )

    result = np.zeros((arr.shape), dtype="float32")

    if channel_last:
        for idx_d in range(arr.shape[2]):
            result[:, :, idx_d] = _convolve_array_2D(
                arr[:, :, idx_d],
                offsets,
                weights,
                method=method,
                nodata=nodata,
                nodata_value=nodata_value,
                func_value=func_value,
            )
    else:
        for idx_d in range(arr.shape[0]):
            result[idx_d, :, :] = _convolve_array_2D(
                arr[idx_d, :, :],
                offsets,
                weights,
                method=method,
                nodata=nodata,
                nodata_value=nodata_value,
                func_value=func_value,
            )

    return result



"""
### Utility functions to work with GDAL ###

These functions are used to interact with basic GDAL objects.
"""

# Standard Library
import sys; sys.path.append("../../")
from typing import Optional, Union, List, Any, Tuple
from warnings import warn
import os

# External
from osgeo import gdal, ogr
import psutil
import numpy as np

# Internal
from buteo.utils import utils_path, utils_translate



def _get_default_creation_options(
    options: Optional[list] = None,
) -> list:
    """
    Takes a list of GDAL creation options and adds the following defaults to it if their not specified: </br>

    Default options are:
    ```python
    >>> default_options = [
    ...     "TILED=YES",
    ...     "NUM_THREADS=ALL_CPUS",
    ...     "BIGTIFF=IF_SAFER",
    ...     "COMPRESS=LZW",
    ...     "BLOCKXSIZE=256",
    ...     "BLOCKYSIZE=256",
    ... ]
    ```

    Parameters
    ----------
    options : Optional[list], optional
        A list of GDAL creation options, default: None

    Returns
    -------
    list
        A list of GDAL creation options with the defaults added.
    """
    assert isinstance(options, (list, type(None))), "Options must be a list or None."

    if options is None:
        options = []

    internal_options = list(options)

    opt_str = " ".join(internal_options)
    if "TILED" not in opt_str:
        internal_options.append("TILED=YES")

    if "NUM_THREADS" not in opt_str:
        internal_options.append("NUM_THREADS=ALL_CPUS")

    if "BIGTIFF" not in opt_str:
        internal_options.append("BIGTIFF=IF_SAFER")

    if "COMPRESS" not in opt_str:
        internal_options.append("COMPRESS=LZW")

    if "BLOCKXSIZE" not in opt_str:
        internal_options.append("BLOCKXSIZE=256")

    if "BLOCKYSIZE" not in opt_str:
        internal_options.append("BLOCKYSIZE=256")

    return internal_options


def get_gdal_memory() -> list:
    """
    Get at list of all active memory layers in GDAL.
    
    Returns
    -------
    list
        A list of all active memory layers in GDAL.
    """
    if hasattr(gdal, "listdir"):
        datasets = []
        for ds in gdal.listdir("/vsimem"):
            name = ds.name
            if name.startswith("/vsimem"):
                datasets.append(name)
            else:
                datasets.append("/vsimem/" + name)
    elif hasattr(gdal, "ReadDir"):
        datasets = ["/vsimem/" + ds for ds in gdal.ReadDir("/vsimem")]
    else:
        warn("Unable to get list of memory datasets.", RuntimeWarning)
        return []

    return datasets


def clear_gdal_memory() -> None:
    """
    Clears all gdal memory.

    Notes
    -----
    This function is not guaranteed to work.
    It is a best effort attempt to clear all gdal memory.
    """
    memory = get_gdal_memory()

    for dataset in memory:
        gdal.Unlink(dataset)

    if len(get_gdal_memory()) != 0:
        for dataset in get_gdal_memory():
            opened = None

            if _check_is_raster(dataset):
                opened = gdal.Open(dataset)
            elif _check_is_raster(dataset):
                opened = ogr.Open(dataset)
            else:
                print(f"Unable to open dataset: {dataset}")
                continue
            driver = opened.GetDriver()
            driver.Delete(dataset)

            opened = None
            driver = None

    if len(get_gdal_memory()) != 0:
        warn("Failed to clear all GDAL memory.", RuntimeWarning)


def _check_is_valid_ext(ext: str) -> bool:
    """
    Check if a file extension has a valid GDAL or OGR driver.

    Parameters
    ----------
    ext : str
        The file extension to check.

    Returns
    -------
    bool
        True if the file extension is a valid GDAL or OGR driver, False otherwise.
    """
    assert isinstance(ext, str), "ext must be a string."
    if len(ext) == 0:
        return False

    if ext in utils_translate._get_valid_driver_extensions():
        return True

    return False


def _check_is_valid_raster_ext(ext: str) -> bool:
    """
    Check if a file extension has a valid GDAL driver.

    Parameters
    ----------
    ext : str
        The file extension to check.

    Returns
    -------
    bool
        True if the file extension is a valid GDAL Raster driver, False otherwise.
    """
    assert isinstance(ext, str), "ext must be a string."
    if len(ext) == 0:
        return False

    if ext in utils_translate._get_valid_raster_driver_extensions():
        return True

    return False


def _check_is_valid_vector_ext(ext: str) -> bool:
    """
    Check if a file extension has a valid OGR driver.

    Parameters
    ----------
    ext : str
        The file extension to check.

    Returns
    -------
    bool
        True if the file extension is a valid OGR Vector driver, False otherwise.
    """
    assert isinstance(ext, str), "ext must be a string."
    if len(ext) == 0:
        return False

    if ext in utils_translate._get_valid_vector_driver_extensions():
        return True

    return False


def _check_is_file_valid_ext(file_path: str) -> bool:
    """
    Check if a file path has a valid GDAL or OGR driver.

    Parameters
    ----------
    file_path : str
        The file path to check.

    Returns
    -------
    bool
        True if the file path is a valid GDAL or OGR driver, False otherwise.
    """
    assert isinstance(file_path, str), "file_path must be a string."
    assert len(file_path) > 0, "file_path cannot be an empty string."

    ext = utils_path._get_ext_from_path(file_path)

    if ext in utils_translate._get_valid_driver_extensions():
        return True

    return False


def _check_is_file_valid_raster_ext(file_path: str) -> bool:
    """
    Check if a file path has a valid GDAL driver.

    Parameters
    ----------
    file_path : str
        The file path to check.

    Returns
    -------
    bool
        True if the file path is a valid GDAL Raster driver, False otherwise.
    """
    assert isinstance(file_path, str), "file_path must be a string."
    assert len(file_path) > 0, "file_path cannot be an empty string."

    ext = utils_path._get_ext_from_path(file_path)

    if ext in utils_translate._get_valid_raster_driver_extensions():
        return True

    return False


def _check_is_file_valid_vector_ext(file_path: str) -> bool:
    """
    Check if a file path has a valid OGR driver.

    Parameters
    ----------
    file_path : str
        The file path to check.

    Returns
    -------
    bool
        True if the file path is a valid OGR Vector driver, False otherwise.
    """
    assert isinstance(file_path, str), "file_path must be a string."
    assert len(file_path) > 0, "file_path cannot be an empty string."

    ext = utils_path._get_ext_from_path(file_path)

    if ext in utils_translate._get_valid_vector_driver_extensions():
        return True

    return False


def _get_default_driver_raster() -> gdal.Driver:
    """
    Get the default GDAL raster driver.

    Returns
    -------
    gdal.Driver
        The default GDAL raster driver.
    """
    return gdal.GetDriverByName("GTiff")


def _get_default_driver_vector() -> ogr.Driver:
    """
    Get the default OGR vector driver.

    Returns
    -------
    ogr.Driver
        The default OGR vector driver.
    """
    return ogr.GetDriverByName("GPKG")


def _get_driver_name_from_path(file_path: str) -> str:
    """
    Convert a file path to a GDAL or OGR driver ShortName (e.g. "GTiff" for "new_york.tif")

    Parameters
    ----------
    file_path : str
        The file path to convert.

    Returns
    -------
    str
        The GDAL or OGR driver ShortName.
    """
    assert isinstance(file_path, str), "file_path must be a string."
    assert len(file_path) > 0, "file_path cannot be an empty string."

    ext = utils_path._get_ext_from_path(file_path)

    if _check_is_file_valid_ext(file_path):
        return utils_translate._get_driver_shortname_from_ext(ext)

    raise ValueError(f"Unable to parse GDAL or OGR driver from path: {file_path}")


def _get_vector_driver_name_from_path(file_path: str) -> str:
    """
    Convert a file path to an OGR driver ShortName (e.g. "FlatGeoBuf" for "new_york.fgb")

    Parameters
    ----------
    file_path : str
        The file path to convert.

    Returns
    -------
    str
        The OGR driver ShortName.
    """
    assert isinstance(file_path, str), "file_path must be a string."
    assert len(file_path) > 0, "file_path cannot be an empty string."

    ext = utils_path._get_ext_from_path(file_path)

    if _check_is_file_valid_vector_ext(file_path):
        return utils_translate._get_vector_shortname_from_ext(ext)

    raise ValueError(f"Unable to parse GDAL or OGR driver from path: {file_path}")


def _get_raster_driver_name_from_path(file_path: str) -> str:
    """
    Convert a file path to a GDAL driver ShortName (e.g. "GTiff" for "new_york.tif")

    Parameters
    ----------
    file_path : str
        The file path to convert.

    Returns
    -------
    str
        The GDAL driver ShortName.
    """
    assert isinstance(file_path, str), "file_path must be a string."
    assert len(file_path) > 0, "file_path cannot be an empty string."

    ext = utils_path._get_ext_from_path(file_path)

    if _check_is_file_valid_raster_ext(file_path):
        return utils_translate._get_raster_shortname_from_ext(ext)

    raise ValueError(f"Unable to parse GDAL or OGR driver from path: {file_path}")


def _check_is_dataset_in_memory(
    raster_or_vector: Union[str, gdal.Dataset, ogr.DataSource],
) -> bool:
    """
    Check if vector is in memory

    Parameters
    ----------
    raster_or_vector : Union[str, gdal.Dataset, ogr.DataSource]
        The raster or vector to check.
    
    Returns
    -------
    bool
        True if the raster or vector is in memory, False otherwise.
    """
    assert isinstance(raster_or_vector, (str, gdal.Dataset, ogr.DataSource)), "raster_or_vector must be a string, gdal.Dataset, or ogr.DataSource."

    if isinstance(raster_or_vector, str):
        if utils_path._check_file_exists_vsimem(raster_or_vector):
            return True

        return False

    elif isinstance(raster_or_vector, (gdal.Dataset, ogr.DataSource)):
        driver = raster_or_vector.GetDriver()
        driver_short_name = None
        try:
            driver_short_name = driver.GetName()
        except AttributeError:
            driver_short_name = driver.ShortName

        if driver_short_name in ["MEM", "Memory", "VirtualMem", "VirtualOGR", "Virtual"]:
            return True

        path = raster_or_vector.GetDescription()
        if utils_path._check_file_exists_vsimem(path):
            return True

        return False

    else:
        raise TypeError("vector_or_raster must be a string, ogr.DataSource, or gdal.Dataset")


def delete_dataset_if_in_memory(
    raster_or_vector: Union[str, gdal.Dataset, ogr.DataSource],
) -> bool:
    """
    Delete raster or vector if it is in memory

    Parameters
    ----------
    raster_or_vector : Union[str, gdal.Dataset, ogr.DataSource]
        The raster or vector to delete.

    Returns
    -------
    bool
        True if the raster or vector was deleted, False otherwise.
    """
    if not isinstance(raster_or_vector, (str, gdal.Dataset, ogr.DataSource)):
        return False

    if not isinstance(raster_or_vector, str):
        path = _get_path_from_dataset(raster_or_vector)
    else:
        path = raster_or_vector

    if _check_is_dataset_in_memory(raster_or_vector):
        if isinstance(raster_or_vector, str):
            gdal.Unlink(raster_or_vector)
        else:
            raster_or_vector.Destroy()
            raster_or_vector = None

            gdal.Unlink(path)

        datasets = get_gdal_memory()
        if path not in datasets:
            return True

    return False


def delete_dataset_if_in_memory_list(
    list_of_raster_or_vectors: List[Union[str, gdal.Dataset, ogr.DataSource]],
) -> bool:
    """
    Deletes a list of raster or vector if they are in memory

    Parameters
    ----------
    list_of_raster_or_vectors : List[Union[str, gdal.Dataset, ogr.DataSource]]
        A list of rasters or vectors to delete.

    Returns
    -------
    bool
        True if any of the rasters or vectors were deleted, False otherwise.
    """
    assert isinstance(list_of_raster_or_vectors, list), "list_of_raster_or_vectors must be a list."

    if len(list_of_raster_or_vectors) == 0:
        return True

    deleted = []
    for raster_or_vector in list_of_raster_or_vectors:
        deleted.append(delete_dataset_if_in_memory(raster_or_vector))

    if all(deleted):
        return True

    return False


def _delete_raster_or_vector(
    raster_or_vector: Union[str, gdal.Dataset, ogr.DataSource],
) -> bool:
    """
    Delete raster or vector. Can be used on both in memory and on disk.

    Parameters
    ----------
    raster_or_vector : Union[str, gdal.Dataset, ogr.DataSource]
        The raster or vector to delete.

    Returns
    -------
    bool
        True if the raster or vector was deleted, False otherwise.
    """
    assert isinstance(raster_or_vector, (str, gdal.Dataset, ogr.DataSource)), "raster_or_vector must be a string, gdal.Dataset, or ogr.DataSource."

    if delete_dataset_if_in_memory(raster_or_vector):
        return True

    driver_shortname = _get_driver_name_from_path(raster_or_vector)
    driver = gdal.GetDriverByName(driver_shortname)
    driver.Delete(raster_or_vector)

    if not utils_path._check_file_exists(raster_or_vector):
        return True

    return False


def _check_is_raster_empty(
    raster: gdal.Dataset,
) -> bool:
    """
    Check if a raster has bands or zero width and zero height.

    Parameters
    ----------
    raster : gdal.Dataset
        The raster to check.

    Returns
    -------
    bool
        True if the raster has bands and width and height greater than zero, False otherwise.
    """
    assert isinstance(raster, gdal.Dataset), "raster must be a gdal.Dataset."

    if raster.RasterCount == 0:
        return True

    if raster.RasterXSize == 0 or raster.RasterYSize == 0:
        return True

    return False


def _check_is_vector_empty(
    vector: ogr.DataSource,
) -> bool:
    """
    Check if a vector has features with geometries

    Parameters
    ----------
    vector : ogr.DataSource
        The vector to check.

    Returns
    -------
    bool
        True if the vector has features with geometries, False otherwise.
    """
    assert isinstance(vector, ogr.DataSource), "vector must be an ogr.DataSource."

    layer_count = vector.GetLayerCount()

    if layer_count == 0:
        return True

    for layer in range(0, layer_count):
        layer = vector.GetLayerByIndex(layer)
        layer.ResetReading()

        if layer.GetFeatureCount() > 0:
            feature_count = layer.GetFeatureCount()

            for feature in range(feature_count):
                feature = layer.GetNextFeature()

                # TODO: Expand this to check geometry.
                if feature is not None:
                    return False

    return True


def _check_is_raster(
    potential_raster: Any,
    empty_is_invalid: bool = True,
) -> bool:
    """
    Checks if a variable is a valid raster.

    Parameters
    ----------
    potential_raster : Union[str, gdal.Dataset]
        The variable to check.

    empty_is_invalid : bool, optional
        If True, an empty raster is considered invalid. The default is True.

    Returns
    -------
    bool
        True if the variable is a valid raster, False otherwise.
    """
    if isinstance(potential_raster, str):
        if not utils_path._check_file_exists(potential_raster):
            return False

        try:
            opened = gdal.Open(potential_raster, 0)
        except RuntimeError:
            return False

        if opened is None:
            return False

        if empty_is_invalid and _check_is_raster_empty(opened):
            return False

        opened = None

        return True

    if isinstance(potential_raster, gdal.Dataset):

        if empty_is_invalid and _check_is_raster_empty(potential_raster):
            return False

        return True

    return False


def _check_is_raster_list(
    potential_raster_list,
    empty_is_invalid=True,
) -> bool:
    """
    Checks if a list of variables are full of valid rasters.

    Parameters
    ----------
    potential_raster_list : List[Union[str, gdal.Dataset]]
        The list of variables to check.

    empty_is_invalid : bool, optional
        If True, an empty raster is considered invalid. Default: True.
    """
    if not isinstance(potential_raster_list, list):
        return False

    if len(potential_raster_list) == 0:
        return False

    for element in potential_raster_list:
        if not _check_is_raster(element, empty_is_invalid=empty_is_invalid):
            return False

    return True


def _check_is_vector(
    potential_vector: List[Any],
    empty_is_invalid: bool = True,
) -> bool:
    """
    Checks if a variable is a valid vector.

    Parameters
    ----------
    potential_vector : Union[str, ogr.DataSource]
        The variable to check.

    empty_is_invalid : bool, optional
        If True, an empty vector is considered invalid. The default is True.

    Returns
    -------
    bool
        True if the variable is a valid vector, False otherwise.
    """
    if isinstance(potential_vector, ogr.DataSource):
        if empty_is_invalid and _check_is_vector_empty(potential_vector):
            print(f"Vector: {potential_vector} was empty.")

            return False

        return True

    if isinstance(potential_vector, str):
        try:
            opened = ogr.Open(potential_vector, 0)
            if opened is None:
                extension = os.path.splitext(potential_vector)[1][1:]

                if extension == "memory" or "mem":
                    driver = ogr.GetDriverByName("Memory")
                    opened = driver.Open(potential_vector)

        except RuntimeError:
            return False

        if isinstance(opened, ogr.DataSource):
            if empty_is_invalid and _check_is_vector_empty(opened):
                return False

            return True

    return False


def _check_is_vector_list(
    potential_vector_list: List[Any],
    empty_is_invalid=True,
) -> bool:
    """
    Checks if a variable is a valid list of vectors.

    Parameters
    ----------
    potential_vector_list : List[any]
        The variable to check.

    empty_is_invalid : bool, optional
        If True, an empty vector is considered invalid. Default: True.

    Returns
    -------
    bool
        True if the variable is a valid list of vectors, False otherwise.
    """
    if not isinstance(potential_vector_list, list):
        return False

    if len(potential_vector_list) == 0:
        return False

    for element in potential_vector_list:
        if not _check_is_vector(element, empty_is_invalid=empty_is_invalid):
            return False

    return True


def _check_is_raster_or_vector(
    potential_raster_or_vector: Any,
    empty_is_invalid: bool = True,
) -> bool:
    """
    Checks if a variable is a valid raster or vector.

    Parameters
    ----------
    potential_raster_or_vector : Union[str, gdal.Dataset, ogr.DataSource]
        The variable to check.

    empty_is_invalid : bool, optional
        If True, an empty raster or vector is considered invalid. The default is True.

    Returns
    -------
    bool
        True if the variable is a valid raster or vector, False otherwise.
    """
    if _check_is_raster(potential_raster_or_vector, empty_is_invalid=empty_is_invalid):
        return True

    if _check_is_vector(potential_raster_or_vector, empty_is_invalid=empty_is_invalid):
        return True

    return False


def _check_is_raster_or_vector_list(
    potential_raster_or_vector_list: List[Any],
    empty_is_invalid: bool = True,
) -> bool:
    """
    Checks if a variable is a valid list of rasters or vectors.

    Parameters
    ----------
    potential_raster_or_vector_list : List[Union[str, gdal.Dataset, ogr.DataSource]]
        The variable to check.

    empty_is_invalid : bool, optional
        If True, an empty raster or vector is considered invalid. Default: True.

    Returns
    -------
    bool
        True if the variable is a valid list of rasters or vectors, False otherwise.
    """
    if not isinstance(potential_raster_or_vector_list, list):
        return False

    if len(potential_raster_or_vector_list) == 0:
        return False

    for element in potential_raster_or_vector_list:
        if not _check_is_raster_or_vector(element, empty_is_invalid=empty_is_invalid):
            return False

    return True


def _get_path_from_dataset(
    dataset: Union[str, gdal.Dataset, ogr.DataSource],
    dataset_type: Optional[bool] = None,
) -> str:
    """
    Gets the path from a datasets. Can be vector or raster, string or opened.

    Parameters
    ----------
    dataset : Union[str, gdal.Dataset, ogr.DataSource]
        The dataset to get the path from.

    dataset_type : Optional[bool], optional
        The type of the dataset. Can be "raster", "vector" or None. If None, the type is guessed. Default: None.
    """
    if (dataset_type == "raster" or dataset_type is None) and _check_is_raster(dataset, empty_is_invalid=False):
        if isinstance(dataset, str):
            raster = gdal.Open(dataset, 0)
        elif isinstance(dataset, gdal.Dataset):
            raster = dataset
        else:
            raise RuntimeError(f"Could not read input raster: {raster}")

        path = raster.GetDescription()
        raster = None

        return utils_path._get_unix_path(path)

    if (dataset_type == "vector" or dataset_type is None) and _check_is_vector(dataset, empty_is_invalid=False):
        if isinstance(dataset, str):
            vector = ogr.Open(dataset, 0)
        elif isinstance(dataset, ogr.DataSource):
            vector = dataset
        else:
            raise RuntimeError(f"Could not read input vector: {vector}")

        path = vector.GetDescription()
        vector = None

        return utils_path._get_unix_path(path)

    raise ValueError("The dataset is not a raster or vector.")


def _get_path_from_dataset_list(
    datasets: List[Union[str, gdal.Dataset, ogr.DataSource]],
    allow_mixed: bool = False,
    dataset_type: Optional[bool] = None,
):
    """
    Gets the paths from a list of datasets.

    Parameters
    ----------
    datasets : List[Union[str, gdal.Dataset, ogr.DataSource]]
        The datasets to get the paths from.

    allow_mixed : bool, optional
        If True, vectors and rasters can be mixed. Default: False.

    dataset_type : Optional[bool], optional
        The type of the datasets. Can be "raster", "vector" or None. If None, the type is guessed. Default: None.

    Returns
    -------
    List[str]
        The paths of the datasets.
    """
    assert isinstance(datasets, list), "The datasets must be a list."
    assert isinstance(dataset_type, (str, type(None))), "The dataset_type must be 'raster', 'vector', or None."

    rasters = False
    vectors = False

    outputs = []
    for dataset in datasets:
        if (dataset_type == "raster" or dataset_type is None) and _check_is_raster(dataset, empty_is_invalid=False):
            dataset_type = "raster"
            rasters = True
        elif (dataset_type == "vector" or dataset_type is None) and _check_is_vector(dataset, empty_is_invalid=False):
            dataset_type = "vector"
            vectors = True
        else:
            raise ValueError(f"The dataset is not a raster or vector. {dataset}")

        if rasters and vectors and not allow_mixed:
            raise ValueError("vectors and rasters are mixed.")

        outputs.append(_get_path_from_dataset(dataset, dataset_type=dataset_type))

    return outputs


def _get_raster_size(
    raster: Union[gdal.Dataset, str],
) -> Tuple[float, float]:
    """
    Gets the size of a raster (xres, yres).

    Parameters
    ----------
    target : Union[gdal.Dataset, str]
        The target to get the size from.

    Returns
    -------
    Tuple[float, float] - xres, yres
    """
    assert _check_is_raster(raster), "The target must be a raster."

    try:
        reference = raster if isinstance(raster, gdal.Dataset) else gdal.Open(raster, 0)
    except RuntimeError:
        raise RuntimeError(f"Could not read input raster: {raster}") from None

    transform = reference.GetGeoTransform()
    reference = None

    x_res = transform[1]
    y_res = abs(transform[5])

    return x_res, y_res


def _get_dynamic_memory_limit(
    proportion: float = 0.8,
    *,
    min_mb: int = 100,
    max_mb: Optional[int] = None,
    available: bool = False,
) -> int:
    """
    Returns a dynamic memory limit taking into account total memory and CPU cores.
    The return is in mbytes. For GDAL.

    The value is interpreted as being in megabytes if the value is less than 10000. For values >=10000, this is interpreted as bytes.

    Parameters
    ----------
    percentage : float, optional.
        The percentage of the total memory to use. Default: 0.8.
    
    min_mb : int, optional.
        The minimum number of megabytes to be returned. Default: 100.
    
    available : bool, optional.
        If True, consider available memory instead of total memory. Default: False.

    Returns
    -------
    int
        The dynamic memory limit in bytes.
    """
    assert isinstance(proportion, (int, float)), "percentage must be an integer."
    assert isinstance(min_mb, int), "min_mb must be an integer."
    assert isinstance(available, bool), "available must be a boolean."
    assert min_mb > 0, "min_mb must be > 0."
    assert proportion > 0.0 and proportion <= 1.0, "percentage must be > 0 and <= 1."

    if available:
        dyn_limit = np.rint(
            (psutil.virtual_memory().available * proportion)  / (1024 ** 2),
        )
    else:
        dyn_limit = np.rint(
            (psutil.virtual_memory().total * proportion)  / (1024 ** 2),
        )

    if dyn_limit < min_mb:
        dyn_limit = min_mb

    if max_mb is not None:
        if dyn_limit > max_mb:
            dyn_limit = max_mb

    # GDALWarpMemoryLimit() expects the value in bytes if it is >= 10000
    if dyn_limit > 10000:
        dyn_limit = dyn_limit * (1024 ** 2)

    return int(dyn_limit)


def _convert_to_band_list(
    band_number: Union[int, List[int]],
    band_count: int,
) -> List[int]:
    """
    Converts a band number or list of band numbers to a list of band numbers.

    Parameters
    ----------
    band_number : Union[int, List[int]]
        The band number or list of band numbers to convert.

    band_count : int
        The number of bands in the raster.

    Returns
    -------
    List[int]
        The list of band numbers.
    """
    if not isinstance(band_number, (int, list)):
        raise TypeError(f"Invalid type for band: {type(band_number)}")
    if not isinstance(band_count, int):
        raise TypeError(f"Invalid type for band_count: {type(band_count)}")

    if band_number == -1:
        band_number = list(range(1, band_count + 1))

    input_list = band_number if isinstance(band_number, list) else [band_number]
    output_list = []

    if band_count <= 0:
        raise ValueError("Band count cannot be 0.")

    if len(input_list) == 0:
        raise ValueError("Band number cannot be 0.")

    if any([val <= 0 or val > band_count for val in input_list]):
        raise ValueError("Band number cannot be 0 or above the band count.")

    for val in input_list:
        output_list.append(int(val))

    return output_list



"""
### Perform morphological operations on arrays and rasters.  ###
"""

# Standard library
import sys; sys.path.append("../../")
from typing import Union


# External
import numpy as np

# Internal
from buteo.array.convolution import convolve_array
from buteo.array.convolution_kernels import kernel_base, kernel_get_offsets_and_weights
from buteo.utils.utils_base import _type_check


def _morphology_operation(
    arr: np.ndarray,
    method: int,
    radius: Union[int, float] = 1,
    spherical: bool = True,
    channel_last: bool = True,
) -> np.ndarray:
    _type_check(arr, [np.ndarray], "arr")
    _type_check(method, [int], "method")
    _type_check(radius, [int, float], "radius")
    _type_check(spherical, [bool], "spherical")
    _type_check(channel_last, [bool], "channel_last")

    kernel = kernel_base(radius, circular=spherical, normalised=False)
    offsets, weights = kernel_get_offsets_and_weights(kernel)

    mask = None
    if np.ma.isMaskedArray(arr):
        nodata = True
        nodata_value = arr.fill_value
        arr = np.ma.getdata(arr)
        mask = np.ma.getmask(arr)
    else:
        nodata = False
        nodata_value = 0.0

    arr_convolved = convolve_array(
        arr,
        offsets,
        weights,
        method=method,
        nodata=nodata,
        nodata_value=nodata_value,
        channel_last=channel_last,
    )

    if nodata:
        arr_convolved = np.ma.array(arr_convolved, mask=mask, fill_value=nodata_value)

    return arr_convolved


def morph_erode(
    arr: np.ndarray,
    radius: Union[int, float] = 1,
    spherical: bool = True,
    channel_last: bool = True,
) -> np.ndarray:
    """
    Erode an array by taking the local minimum.
    
    Parameters
    ----------
    arr : np.ndarray
        Input array.

    radius : Union[int, float], optional
        Radius of the kernel. Default: 1 (3x3 kernel)

    spherical : bool, optional
        Use a spherical kernel. Default: True

    channel_last : bool, optional
        If True, the channel axis is the last axis. Default: True

    Returns
    -------
    np.ndarray
        Eroded array.
    """
    return _morphology_operation(
        arr,
        method="min",
        radius=radius,
        spherical=spherical,
        channel_last=channel_last,
    )


def morph_dilate(
    arr: np.ndarray,
    radius: Union[int, float] = 1,
    spherical: bool = True,
    channel_last: bool = True,
) -> np.ndarray:
    """
    Dilate an array by taking the local maximum.
    
    Parameters
    ----------
    arr : np.ndarray
        Input array.

    radius : Union[int, float], optional
        Radius of the kernel. Default: 1 (3x3 kernel)

    spherical : bool, optional
        Use a spherical kernel. Default: True

    channel_last : bool, optional
        If True, the channel axis is the last axis. Default: True

    Returns
    -------
    np.ndarray
        Dilated array.
    """
    return _morphology_operation(
        arr,
        method="max",
        radius=radius,
        spherical=spherical,
        channel_last=channel_last,
    )


def morph_open(
    arr: np.ndarray,
    radius: Union[int, float] = 1,
    spherical: bool = True,
    channel_last: bool = True,
) -> np.ndarray:
    """
    Perform the open mortholigical operation on an array.
    
    Parameters
    ----------
    arr : np.ndarray
        Input array.

    radius : Union[int, float], optional
        Radius of the kernel. Default: 1 (3x3 kernel)

    spherical : bool, optional
        Use a spherical kernel. Default: True

    channel_last : bool, optional
        If True, the channel axis is the last axis. Default: True

    Returns
    -------
    np.ndarray
        Opened array.
    """
    erode = morph_erode(
        arr,
        radius=radius,
        spherical=spherical,
        channel_last=channel_last,
    )

    return morph_dilate(
        erode,
        radius=radius,
        spherical=spherical,
        channel_last=channel_last,
    )


def morph_close(
    arr: np.ndarray,
    radius: Union[int, float] = 1,
    spherical: bool = True,
    channel_last: bool = True,
) -> np.ndarray:
    """
    Perform the close morphological operation on an array.

    Parameters
    ----------
    arr : np.ndarray
        Input array.

    radius : Union[int, float], optional
        Radius of the kernel. Default: 1 (3x3 kernel)

    spherical : bool, optional
        Use a spherical kernel. Default: True

    channel_last : bool, optional
        If True, the channel axis is the last axis. Default: True

    Returns
    -------
    np.ndarray
        Closed array.
    """
    dilate = morph_dilate(
        arr,
        radius=radius,
        spherical=spherical,
        channel_last=channel_last,
    )

    return morph_erode(
        dilate,
        radius=radius,
        spherical=spherical,
        channel_last=channel_last,
    )


def morph_tophat(
    arr: np.ndarray,
    radius: Union[int, float] = 1,
    spherical: bool = True,
    channel_last: bool = True,
) -> np.ndarray:
    """
    Perform the top_hat morphological operation on an array.

    Same as: `array / opened(array)`

    Parameters
    ----------
    arr : np.ndarray
        Input array.

    radius : Union[int, float], optional
        Radius of the kernel. Default: 1 (3x3 kernel)

    spherical : bool, optional
        Use a spherical kernel. Default: True

    channel_last : bool, optional
        If True, the channel axis is the last axis. Default: True

    Returns
    -------
    np.ndarray
        TopHat array.
    """
    opened = morph_open(
        arr,
        radius=radius,
        spherical=spherical,
        channel_last=channel_last,
    )

    return arr / opened


def morph_bothat(
    arr: np.ndarray,
    radius: Union[int, float] = 1,
    spherical: bool = True,
    channel_last: bool = True,
) -> np.ndarray:
    """
    Perform the bottom_hat morphological operation on an array.
    
    Same as: `closed(array) / array`

    Parameters
    ----------
    arr : np.ndarray
        Input array.

    radius : Union[int, float], optional
        Radius of the kernel. Default: 1 (3x3 kernel)

    spherical : bool, optional
        Use a spherical kernel. Default: True

    channel_last : bool, optional
        If True, the channel axis is the last axis. Default: True

    Returns
    -------
    np.ndarray
        BotHat array.
    """
    closed = morph_close(
        arr,
        radius=radius,
        spherical=spherical,
        channel_last=channel_last,
    )

    return closed / arr


def morph_difference(
    arr: np.ndarray,
    radius: Union[int, float] = 1,
    spherical: bool = True,
    channel_last: bool = True,
) -> np.ndarray:
    """
    Perform the difference morphological operation on an array.
    
    Same as: `dilate(array) - erode(array)`

    Parameters
    ----------
    arr : np.ndarray
        Input array.

    radius : Union[int, float], optional
        Radius of the kernel. Default: 1 (3x3 kernel)

    spherical : bool, optional
        Use a spherical kernel. Default: True

    channel_last : bool, optional
        If True, the channel axis is the last axis. Default: True

    Returns
    -------
    np.ndarray
        Difference array.
    """
    erode = morph_erode(
        arr,
        radius=radius,
        spherical=spherical,
        channel_last=channel_last,
    )
    dilate = morph_dilate(
        arr,
        radius=radius,
        spherical=spherical,
        channel_last=channel_last,
    )

    return dilate - erode



"""
### Generic utility functions ###

Functions that make interacting with the toolbox easier.
"""
# Standard library
from typing import Union, Optional
import sys; sys.path.append("../../")

# External
import numpy as np
from numba import jit, prange

# Internal
from buteo.array.utils_array import _create_grid


# TODO: Multi-channel support, Split assert
@jit(nopython=True, parallel=True, fastmath=True, cache=True, nogil=True)
def convolve_distance(
    array: np.ndarray,
    target: Union[int, float] = 1,
    maximum_distance: Optional[Union[int, float]] = None,
    pixel_width: Union[int, float] = 1,
    pixel_height: Union[int, float] = 1,
) -> np.ndarray:
    """
    Calculate the distance from each pixel to the nearest target pixel.
    
    Parameters
    ----------
    array : np.ndarray
        The array to calculate the distances for.

    target : Union[int, float], optional
        The target value to calculate the distance to. Default: 1.

    maximum_distance : Union[int, float], optional
        The maximum distance to calculate. Default: None.

    pixel_width : Union[int, float], optional
        The width of each pixel. Default: 1.

    pixel_height : Union[int, float], optional
        The height of each pixel. Default: 1.

    Returns
    -------
    np.ndarray
        The array of distances.
    """
    binary_array = np.sum(array == target, axis=2, dtype=np.uint8)

    if maximum_distance is None:
        maximum_distance = np.sqrt(binary_array.shape[0] ** 2 + binary_array.shape[1] ** 2)

    radius_cols = int(np.ceil(maximum_distance / pixel_height))
    radius_rows = int(np.ceil(maximum_distance / pixel_width))

    kernel_cols = radius_cols * 2
    kernel_rows = radius_rows * 2

    if kernel_cols % 2 == 0:
        kernel_cols += 1

    if kernel_rows % 2 == 0:
        kernel_rows += 1

    middle_cols = int(np.floor(kernel_cols / 2))
    middle_rows = int(np.floor(kernel_rows / 2))

    range_cols = np.arange(-middle_cols, middle_cols + 1)
    range_rows = np.arange(-middle_rows, middle_rows + 1)

    cols_grid, rows_grid = _create_grid(range_rows, range_cols)
    coord_grid = np.empty((cols_grid.size, 2), dtype=np.int64)
    coord_grid[:, 0] = cols_grid.flatten()
    coord_grid[:, 1] = rows_grid.flatten()

    coord_grid_projected = np.empty_like(coord_grid, dtype=np.float32)
    coord_grid_projected[:, 0] = coord_grid[:, 0] * pixel_height
    coord_grid_projected[:, 1] = coord_grid[:, 1] * pixel_width

    coord_grid_values = np.sqrt((coord_grid_projected[:, 0] ** 2) + (coord_grid_projected[:, 1] ** 2))

    selected_range = np.arange(coord_grid.shape[0])
    selected_range = selected_range[np.argsort(coord_grid_values)][1:]
    selected_range = selected_range[coord_grid_values[selected_range] <= maximum_distance]

    coord_grid = coord_grid[selected_range]
    coord_grid_values = coord_grid_values[selected_range]

    distances = np.full_like(binary_array, maximum_distance, dtype=np.float32)
    for col in prange(binary_array.shape[0]):
        for row in range(binary_array.shape[1]):
            if binary_array[col, row] == target:
                distances[col, row] = 0
            else:
                for idx, (col_adj, row_adj) in enumerate(coord_grid):
                    if (col + col_adj) >= 0 and (col + col_adj) < binary_array.shape[0] and \
                        (row + row_adj) >= 0 and (row + row_adj) < binary_array.shape[1] and \
                        binary_array[col + col_adj, row + row_adj] == target:

                        distances[col, row] = coord_grid_values[idx]
                        break

    distance_array = np.expand_dims(distances, axis=2)

    return distance_array



"""
### Reproject vectors. ###

Functions to reproject vectors. References can be both vector and raster.
"""

# Standard library
import sys; sys.path.append("../../")
from typing import Union, Optional, List

# External
from osgeo import gdal, osr, ogr

# Internal
from buteo.utils import (
    utils_io,
    utils_gdal,
    utils_base,
    utils_path,
    utils_projection,
)


def _vector_reproject(
    vector: Union[str, ogr.DataSource],
    projection: Union[str, int, osr.SpatialReference, gdal.Dataset, ogr.DataSource],
    out_path: Optional[str] = None,
    copy_if_same: bool = False,
    prefix: str = "",
    suffix: str = "",
    add_uuid: bool = False,
):
    """ Internal. """
    assert isinstance(vector, (ogr.DataSource, str)), "Invalid vector input"
    assert utils_gdal._check_is_vector(vector), "Invalid vector input"

    # The input is already in the correct projection.
    if not copy_if_same:
        original_projection = utils_projection.parse_projection(vector)

        if utils_projection._check_projections_match(original_projection, projection):
            return utils_gdal._get_path_from_dataset(vector)

    in_path = utils_gdal._get_path_from_dataset(vector)

    if out_path is None:
        out_path = utils_path._get_temp_filepath(vector, add_uuid=add_uuid, prefix=prefix, suffix=suffix)

    options = []
    wkt = utils_projection.parse_projection(projection, return_wkt=True).replace(" ", "\\")

    options.append(f'-t_srs "{wkt}"')

    success = gdal.VectorTranslate(
        out_path,
        in_path,
        format=utils_gdal._get_vector_driver_name_from_path(out_path),
        options=" ".join(options),
    )

    if success != 0 and success is not None:
        return out_path
    else:
        raise RuntimeError("Error while reprojecting geometry.")


def vector_reproject(
    vector: Union[Union[str, ogr.DataSource], List[Union[str, ogr.DataSource]]],
    projection: Union[str, int, osr.SpatialReference, gdal.Dataset, ogr.DataSource],
    out_path: Optional[str] = None,
    copy_if_same: bool = False,
    prefix: str = "",
    suffix: str = "",
    add_uuid: bool = False,
    overwrite: bool = True,
):
    """
    Reprojects a vector given a target projection.

    Parameters
    ----------
    vector : Union[str, ogr.DataSource]
        The vector to reproject.

    projection : Union[str, int, osr.SpatialReference, gdal.Dataset, ogr.DataSource]
        The projection is infered from the input. The input can be: WKT proj, EPSG proj, Proj, or read from a vector or raster datasource either from path or in-memory.

    out_path : Optional[str], optional
        The destination to save to. If None then the output is an in-memory raster., default: None

    copy_if_same : bool, optional
        Create a copy, even if the projections are the same., default: False

    prefix : str, optional
        The prefix to add to the output path., default: ""

    suffix : str, optional
        The suffix to add to the output path., default: ""

    add_uuid : bool, optional
        Add a uuid to the output path., default: False

    overwrite : bool, optional
        Is it possible to overwrite the out_path if it exists., default: True

    Returns
    -------
    Union[str, List[str]]
        An in-memory vector. If an out_path is given, the output is a string containing the path to the newly created vecotr.
    """
    utils_base._type_check(vector, [str, ogr.DataSource], "vector")
    utils_base._type_check(projection, [str, int, ogr.DataSource, gdal.Dataset, osr.SpatialReference], "projection")
    utils_base._type_check(out_path, [str, [str], None], "out_path")
    utils_base._type_check(copy_if_same, [bool], "copy_if_same")
    utils_base._type_check(prefix, [str], "prefix")
    utils_base._type_check(suffix, [str], "suffix")
    utils_base._type_check(add_uuid, [bool], "add_uuid")
    utils_base._type_check(overwrite, [bool], "overwrite")

    input_is_list = isinstance(vector, list)

    input_data = utils_io._get_input_paths(vector, "vector")
    output_data = utils_io._get_output_paths(
        input_data,
        out_path,
        prefix=prefix,
        suffix=suffix,
        add_uuid=add_uuid,
        overwrite=overwrite,
    )

    utils_path._delete_if_required_list(output_data, overwrite)

    output = []
    for idx, in_vector in enumerate(input_data):
        output.append(
            _vector_reproject(
                in_vector,
                projection,
                out_path=output_data[idx],
                prefix=prefix,
                suffix=suffix,
                add_uuid=add_uuid,
                copy_if_same=copy_if_same,
            )
        )

    if input_is_list:
        return output

    return output[0]



"""
### Perform morphological operations on arrays and rasters.  ###
"""

# Standard library
import sys; sys.path.append("../../")
from typing import Union

# External
import numpy as np

# Internal
from buteo.array.convolution import convolve_array, pad_array
from buteo.array.convolution_kernels import kernel_sobel, kernel_get_offsets_and_weights


def filter_edge_detection(
    arr: np.ndarray,
    radius: Union[int, float] = 1,
    scale: Union[int, float] = 1,
    gradient: bool = False,
    channel_last: bool = True,
) -> np.ndarray:
    """
    Perform edge detection on an array using a Sobel-style operator.

    Parameters
    ----------
    arr : np.ndarray
        Input array.

    radius : Union[int, float], optional
        Radius of the kernel. Default: 1 (3x3 kernel)
        The radius can be fractional.

    scale : Union[int, float], optional
        scale of the edge detection. Increase to increase the
        sensitivity of the edge detection. This can also be fractional.
        Default: 1.0

    gradient : bool, optional
        Return the gradient as well as the magnitude, default: False.

    Returns
    -------
    np.ndarray
        Edge detection result.
    """
    arr = arr.astype(np.float32, copy=False)
    mask = None
    if np.ma.isMaskedArray(arr):
        nodata = True
        nodata_value = arr.fill_value
        arr = np.ma.getdata(arr)
        mask = np.ma.getmask(arr)
    else:
        nodata = False
        nodata_value = 0.0

    kernel_gx, kernel_gy = kernel_sobel(radius, scale)

    offsets_gx, weights_gx = kernel_get_offsets_and_weights(kernel_gx)
    offsets_gy, weights_gy = kernel_get_offsets_and_weights(kernel_gy)

    pad_size = kernel_gx.shape[0] // 2
    padded = pad_array(arr, pad_size)

    arr_gx = convolve_array(
        padded,
        offsets_gx,
        weights_gx,
        nodata=nodata,
        nodata_value=nodata_value,
        channel_last=channel_last,
    )

    arr_gy = convolve_array(
        padded,
        offsets_gy,
        weights_gy,
        nodata=nodata,
        nodata_value=nodata_value,
        channel_last=channel_last,
    )

    magnitude = np.sqrt(arr_gx ** 2 + arr_gy ** 2)
    magnitude = magnitude[pad_size:-pad_size, pad_size:-pad_size]

    if nodata:
        magnitude = np.ma.array(magnitude, mask=mask, fill_value=nodata_value)

    if gradient:
        gradient = np.arctan2(arr_gy, arr_gx)
        gradient = gradient[pad_size:-pad_size, pad_size:-pad_size]

        return magnitude, gradient

    return magnitude



"""
Work with vector data and attributes with ease.
"""
from .core_vector import *
from .clip import *
from .convert_parts import *
from .dissolve import *
from .grid import *
from .intersect import *
from .merge import *
from .rasterize import *
from .reproject import *
from .buffer import *
from .split import *
from .metadata import *
from .shapes import *



"""
### Read metadata from rasters ###
"""
# Standard library
import sys; sys.path.append("../../")
from typing import List, Union

# External
from osgeo import gdal

# Internal
from buteo.utils import (
    utils_io,
    utils_base,
)
from buteo.raster.core_raster import _get_basic_metadata_raster


def _raster_to_metadata(
    raster: Union[str, gdal.Dataset],
) -> dict:
    """ Internal. """
    utils_base._type_check(raster, [str, gdal.Dataset], "raster")

    metadata = _get_basic_metadata_raster(raster)

    return metadata


def raster_to_metadata(
    raster: Union[str, gdal.Dataset, List[str], List[gdal.Dataset]],
) -> Union[dict, List[dict]]:
    """
    Reads metadata from a raster dataset or a list of raster datasets, and returns a dictionary or a list of dictionaries
    containing metadata information for each raster.

    Bounding boxes are in the format `[xmin, xmax, ymin, ymax]`. These are the keys in the dictionary:

    - `path` (str): Path to the raster.
    - `basename` (str): Basename of the raster.
    - `name` (str): Name of the raster without extension.
    - `folder` (str): Folder of the raster.
    - `ext` (str): Extension of the raster.
    - `in_memory` (bool): Whether the raster is in memory or not.
    - `driver` (str): Driver of the raster.
    - `projection_osr` (osr.SpatialReference): Projection of the raster as an osr.SpatialReference object.	
    - `projection_wkt` (str): Projection of the raster as a WKT string.
    - `geotransform` (tuple): Geotransform of the raster.
    - `size` (tuple): Size of the raster in pixels.
    - `shape` (list): Shape of the raster in pixels. (height, width, bands)
    - `height` (int): Height of the raster in pixels.
    - `width` (int): Width of the raster in pixels.
    - `pixel_size` (tuple): Pixel size of the raster in units of the projection.
    - `pixel_width` (float): Pixel width of the raster in units of the projection.
    - `pixel_height` (float): Pixel height of the raster in units of the projection.
    - `origin` (tuple): Origin of the raster in units of the projection.
    - `origin_x` (float): Origin x of the raster in units of the projection.
    - `origin_y` (float): Origin y of the raster in units of the projection.
    - `bbox` (list): Bounding box of the raster in units of the projection. `[xmin, xmax, ymin, ymax]`
    - `bbox_gdal` (list): Bounding box of the raster in GDAL format. `[xmin, ymin, xmax, ymax]`
    - `bbox_latlng` (list): Bounding box of the raster in latlng format. `[ymin, xmin, ymax, xmax]`
    - `bounds_latlng` (str): Bounding box of the raster in latlng format as a WKT string.
    - `x_min` (float): Minimum x of the raster in units of the projection.
    - `x_max` (float): Maximum x of the raster in units of the projection.
    - `y_min` (float): Minimum y of the raster in units of the projection.
    - `y_max` (float): Maximum y of the raster in units of the projection.
    - `bands` (int): Number of bands in the raster.
    - `dtype_gdal` (int): GDAL data type of the raster.
    - `dtype` (numpy.dtype): Numpy data type of the raster.
    - `dtype_name` (str): Name of the numpy data type of the raster.
    - `area_latlng` (float): Area of the raster in latlng units.
    - `area` (float): Area of the raster in units of the projection.

    Parameters
    ----------
    raster : str or gdal.Dataset or list
        A path to a raster or a gdal.Dataset, or a list of paths to rasters.

    Returns
    -------
    dict or list of dict
        A dictionary or a list of dictionaries containing metadata information for each raster.
    """
    utils_base._type_check(raster, [str, gdal.Dataset, [str, gdal.Dataset]], "raster")

    input_is_list = isinstance(raster, list)
    input_rasters = utils_io._get_input_paths(raster, "raster")

    metadata = []
    for in_raster in input_rasters:
        metadata.append(_raster_to_metadata(in_raster))

    if input_is_list:
        return metadata

    return metadata[0]



"""
### Vectorize rasters. ###

Module to turn rasters into vector representations.
"""

# Standard library
import sys; sys.path.append("../../")
from typing import Union, Optional, List
import os

# External
from osgeo import gdal, ogr

# Internal
from buteo.utils import (
    utils_base,
    utils_gdal,
    utils_path,
    utils_io
)
from buteo.raster import core_raster



def _raster_vectorize(
    raster: Union[str, gdal.Dataset],
    out_path: Optional[str] = None,
    band: int = 1,
    attribute_field: Optional[str] = None,
):
    """ Internal. """
    meta = core_raster._get_basic_metadata_raster(raster)
    opened = core_raster._raster_open(raster)
    src_band = opened.GetRasterBand(band)

    projection = meta["projection_osr"]

    if out_path is None:
        out_path = utils_path._get_temp_filepath("vectorized_raster.gpkg")
    else:
        if not utils_path._check_is_valid_output_filepath(out_path, overwrite=True):
            raise ValueError(f"out_path is not a valid output path: {out_path}")

    utils_path._delete_if_required(out_path, overwrite=True)

    driver_name = utils_gdal._get_vector_driver_name_from_path(out_path)
    driver = ogr.GetDriverByName(driver_name)

    datasource = driver.CreateDataSource(out_path)
    layer_name = os.path.basename(utils_path._get_filename_from_path(out_path))
    layer = datasource.CreateLayer(layer_name, srs=projection)

    if attribute_field is None:
        attr_index = -1
    else:
        metadata = core_raster._get_basic_metadata_raster(raster)
        layer = metadata["layers"][0]["field_names"]
        if attribute_field not in layer:
            raise AttributeError(f"Attribute field ({attribute_field}) not found in raster.")

        attr_index = layer.index(attribute_field)

    try:
        gdal.Polygonize(src_band, None, layer, attr_index)
    except:
        raise RuntimeError(f"Error while vectorizing raster. {raster}") from None
    finally:
        datasource.FlushCache()
        src_band = None
        datasource = None

    return out_path


def raster_vectorize(
    raster: Union[str, gdal.Dataset, List],
    out_path: Optional[str] = None,
    band: int = 1,
    attribute_field: Optional[str] = None,
    prefix: str = "",
    suffix: str = "",
    add_uuid: bool = False,
    add_timestamp: bool = False,
    overwrite: bool = True,
) -> Union[str, List[str]]:
    """
    Vectorizes a raster by turning it into polygons per unique value. Works
    best on integer rasters.

    Parameters
    ----------
    raster : Union[str, gdal.Dataset, List]
        The raster(s) to vectorize.

    out_path : Optional[str], optional
        The path(s) to save the vectorized raster(s) to. Default: None
    
    band : int, optional
        The band to vectorize. Default: 1

    prefix : str, optional
        The prefix to be added to the output raster name. Default: ""

    suffix : str, optional
        The suffix to be added to the output raster name. Default: ""

    add_uuid : bool, optional
        If True, a unique identifier will be added to the output raster name. Default: False

    add_timestamp : bool, optional
        If True, a timestamp will be added to the output raster name. Default: False

    overwrite : bool, optional
        If True, the output raster will be overwritten if it already exists. Default: True

    Returns
    -------
    Union[str, List]
        The path(s) to the vectorized raster(s).
    """
    utils_base._type_check(raster, [str, gdal.Dataset, [str, gdal.Dataset]], "raster")
    utils_base._type_check(out_path, [str, [str], None], "out_path")
    utils_base._type_check(band, [int], "band")
    utils_base._type_check(attribute_field, [str, None], "attribute_field")
    utils_base._type_check(prefix, [str], "prefix")
    utils_base._type_check(suffix, [str], "suffix")
    utils_base._type_check(add_uuid, [bool], "add_uuid")
    utils_base._type_check(add_timestamp, [bool], "add_timestamp")
    utils_base._type_check(overwrite, [bool], "overwrite")

    input_is_list = isinstance(raster, list)

    input_list = utils_io._get_input_paths(raster, "raster")
    output_list = utils_io._get_output_paths(
        input_list,
        out_path,
        overwrite=overwrite,
        prefix=prefix,
        suffix=suffix,
        add_uuid=add_uuid,
        add_timestamp=add_timestamp,
        change_ext="gpkg",
    )

    utils_path._delete_if_required_list(output_list, overwrite)

    vectorized_rasters = []
    for idx, in_raster in enumerate(input_list):
        vectorized_rasters.append(
            _raster_vectorize(
                in_raster,
                out_path=output_list[idx],
                band=band,
                attribute_field=attribute_field,
            )
        )

    if input_is_list:
        return vectorized_rasters

    return vectorized_rasters[0]



"""
This module contains functions for augmenting images that are
suited to remote sensing imagery.
"""
# Standard library
import sys; sys.path.append("../../")
from typing import Optional, List, Callable, Union

# External
import numpy as np

# Buteo
from buteo.array.loaders import MultiArray


class Dataset:
    """
    A dataset that does not apply any augmentations to the data.
    Allows a callback to be passed and can convert between
    channel formats.

    Parameters
    ----------
    X : np.ndarray or MultiArray
        The data to read.

    y : np.ndarray or MultiArray
        The labels for the data.

    callback : callable, optional
        A callback to apply to the data before returning.

    Returns
    -------
    Dataset
        A dataset to iterate over combining X and y.
    """
    def __init__(self,
        X: Union[np.ndarray, MultiArray],
        y: Union[np.ndarray, MultiArray],
        callback: Callable = None,
    ):
        assert len(X) == len(y), "X and y must have the same length."
        assert callback is None or callable(callback), "Callback must be callable."

        self.x_train = X
        self.y_train = y
        self.callback = callback

    def __getitem__(self, index):
        sample_x, sample_y = (self.x_train[index], self.y_train[index])
        
        sample_x = np.array(sample_x) if isinstance(sample_x, np.memmap) else sample_x
        sample_y = np.array(sample_y) if isinstance(sample_y, np.memmap) else sample_y

        # Apply callback if specified
        if self.callback is not None:
            sample_x, sample_y = self.callback(sample_x, sample_y)

        return sample_x, sample_y

    def __len__(self):
        return len(self.x_train)

    def __iter__(self):
        for i in range(len(self)):
            yield self[i]


class DatasetAugmentation:
    """
    A dataset that does not apply any augmentations to the data.
    Allows a callback to be passed and can convert between
    channel formats.

    Parameters
    ----------
    X : np.ndarray or MultiArray
        The data to read.

    y : np.ndarray or MultiArray
        The labels for the data.

    callback_pre_augmentation : callable, optional
        A callback to apply before augmentation.

    callback_post_augmentation : callable, optional
        A callback to apply after augmentation.

    augmentations : list, optional
        The augmentations to apply.
        
    Returns
    -------
    Dataset
        A dataset to iterate over combining X and y.
    """
    def __init__(self,
        X: Union[np.ndarray, MultiArray],
        y: Union[np.ndarray, MultiArray],
        callback_pre_augmentation: Callable = None,
        callback_post_augmentation: Callable = None,
        augmentations: Optional[List] = None,
    ):
        assert len(X) == len(y), "X and y must have the same length."
        assert callback_pre_augmentation is None or callable(callback_pre_augmentation), "callback_pre_augmentation must be callable."
        assert callback_post_augmentation is None or callable(callback_post_augmentation), "callback_post_augmentation must be callable."

        if augmentations is not None:
            assert isinstance(augmentations, list), "Augmentations must be a list."
            self.augmentations = augmentations
        else:
            self.augmentations = []

        self.x_train = X
        self.y_train = y
        self.callback_pre_augmentation = callback_pre_augmentation
        self.callback_post_augmentation = callback_post_augmentation

    def __getitem__(self, index):
        sample_x, sample_y = (self.x_train[index], self.y_train[index])
        
        sample_x = np.array(sample_x) if isinstance(sample_x, np.memmap) else sample_x
        sample_y = np.array(sample_y) if isinstance(sample_y, np.memmap) else sample_y

        # Apply callback if specified
        if self.callback_pre_augmentation is not None:
            sample_x, sample_y = self.callback_pre_augmentation(sample_x, sample_y)

        # Apply augmentations
        for aug in self.augmentations:
            if not isinstance(aug, Callable):
                raise ValueError("Augmentations must be callable.")

            if aug.requires_dataset:
                source_idx = np.random.randint(len(self.x_train))
                source_x, source_y = (self.x_train[source_idx], self.y_train[source_idx])
                
                if self.callback_pre_augmentation is not None:
                    source_x, source_y = self.callback_pre_augmentation(source_x, source_y)

                sample_x, sample_y = aug(sample_x, sample_y, source_x, source_y)

            elif aug.applies_to_features and aug.applies_to_labels:
                sample_x, sample_y = aug(sample_x, sample_y)
            elif aug.applies_to_features:
                sample_x = aug(sample_x)
            elif aug.applies_to_labels:
                sample_y = aug(sample_y)

        # Apply callback if specified
        if self.callback_post_augmentation is not None:
            sample_x, sample_y = self.callback_post_augmentation(sample_x, sample_y)

        return sample_x, sample_y

    def __len__(self):
        return len(self.x_train)

    def __iter__(self):
        for i in range(len(self)):
            yield self[i]



"""
### Utility functions read and write buteo files ###
"""

# Standard Library
import sys; sys.path.append("../../")
from typing import Optional, Union, List

# External
from osgeo import gdal, ogr

# Internal
from buteo.utils import utils_path, utils_base, utils_gdal



def _get_input_paths(
    inputs: Union[gdal.Dataset, ogr.DataSource, str, List[Union[gdal.Dataset, ogr.DataSource, str]]],
    input_type: str,
) -> List[str]:
    """
    Parses the input data to a list of paths.

    Parameters
    ----------
    input_data : Union[gdal.Dataset, ogr.DataSource, str, List[Union[gdal.Dataset, ogr.DataSource, str]]]
        The input data to parse.

    input_type : str, optional
        The input type. Can be "raster", "vector" or "mixed". Default: "mixed".

    Returns
    -------
    List[str]
        The list of paths.
    """
    assert input_type in ["raster", "vector", "mixed"], "Invalid input type."
    assert isinstance(inputs, (gdal.Dataset, ogr.DataSource, str, list)), "Invalid type for input data."

    if isinstance(inputs, str):
        if utils_path._check_is_path_glob(inputs):
            inputs = utils_path._get_paths_from_glob(inputs)
        else:
            inputs = utils_base._get_variable_as_list(inputs)

    elif isinstance(inputs, (gdal.Dataset, ogr.DataSource)):
        inputs = [inputs.GetDescription()]

    # Input is list
    elif isinstance(inputs, list):
        if len(inputs) == 0:
            raise ValueError("Input data cannot be empty.")

        if not all([isinstance(val, (gdal.Dataset, ogr.DataSource, str)) for val in inputs]):
            raise TypeError("Invalid type for input data.")

        inputs = utils_gdal._get_path_from_dataset_list(inputs, allow_mixed=True)

        if input_type == "mixed":
            if not all([utils_gdal._check_is_file_valid_ext(val) for val in inputs]):
                raise TypeError("Invalid type for input data.")
        elif input_type == "raster":
            if not all([utils_gdal._check_is_raster(val) for val in inputs]):
                raise TypeError("Invalid type for input data.")
        elif input_type == "vector":
            if not all([utils_gdal._check_is_vector(val) for val in inputs]):
                raise TypeError("Invalid type for input data.")

    else:
        raise TypeError("Invalid type for input data.")

    if not utils_path._check_is_valid_filepath_list(inputs):
        raise ValueError(f"Unable to locate input data. {str(inputs)}")

    inputs = [utils_path._parse_path(val) for val in inputs]

    for val in inputs:
        if not utils_path._check_file_exists(val):
            raise ValueError(f"Unable to locate input data. {str(val)}")

    if input_type == "raster":
        for val in inputs:
            if not utils_gdal._check_is_raster(val):
                raise TypeError("Invalid raster type for input data.")

    elif input_type == "vector":
        for val in inputs:
            if not utils_gdal._check_is_vector(val):
                raise TypeError("Invalid vector type for input data.")

    else:
        for val in inputs:
            if not utils_gdal._check_is_raster_or_vector(val):
                raise TypeError("Invalid type for input data.")

    return inputs



def _get_output_paths(
    inputs: Union[str, gdal.Dataset, ogr.DataSource, List[Union[str, gdal.Dataset, ogr.DataSource]]],
    output_path: Optional[str] = None,
    in_place: bool = False,
    *,
    prefix: str = "",
    suffix: str = "",
    change_ext: Optional[str] = None,
    add_uuid: bool = False,
    add_timestamp: bool = False,
    overwrite: bool = True,
) -> List[str]:
    """
    Get the output path for a file using the input path and the output path.
    The output path can be None, in which case the created path will be in memory.
    Augmentations can be made to the output.

    Parameters
    ----------
    input_path: str
        The path to the input file.

    output_path: str or None. Optional.
        The path to the output file. If None, the output will be in memory.
        Default: None.

    in_place: bool. Optional.
        If True, the output will be the same as the input. Default: False.

    prefix: str. Optional.
        The prefix to add to the path. Default: "".

    suffix: str. Optional.
        The suffix to add to the path. Default: "".

    change_ext: str. Optional.
        The extension to change the file to. Default: None.

    add_uuid: bool. Optional.
        If True, add a uuid the path. Default: False.

    add_timestamp: bool. Optional.
        If True, add a timestamp the path. Default: False.

    overwrite: bool. Optional.
        If True, if the output path exists and overwrite is False, an error will be thrown.
        Default: True.

    Returns
    -------
    List[str]
        The output path(s).
    """
    assert isinstance(inputs, (str, gdal.Dataset, ogr.DataSource, list)), "input_path must be a string or a list."
    assert isinstance(output_path, (str, type(None), list)), "output_path must be a string or None."
    assert isinstance(in_place, bool), "in_place must be a bool."
    assert isinstance(prefix, str), "prefix must be a string."
    assert isinstance(suffix, str), "suffix must be a string."
    assert isinstance(overwrite, bool), "overwrite must be a bool."
    assert isinstance(add_uuid, bool), "add_uuid must be a bool."
    assert change_ext is None or isinstance(change_ext, str), "change_ext must be a string."

    if output_path is not None:
        assert isinstance(output_path, (str, list)), "output_path must be a string or list."
        if isinstance(output_path, str):
            assert len(output_path) > 0, "output_path must not be empty string."
        elif isinstance(output_path, list):
            assert len(output_path) > 0, "output_path must not be empty list."
            assert all([isinstance(val, str) for val in output_path]), "output_path must be a list of strings."
            assert all([len(val) > 0 for val in output_path]), "output_path must not be a list of empty strings."
            assert len(output_path) == len(inputs), "output_path must be the same length as input_path."

    inputs = _get_input_paths(inputs, input_type="mixed")
    outputs = []

    # In-place
    if in_place:
        for path in inputs:
            outputs.append(path)

    # Output is None - Memory is used.
    elif output_path is None:
        for path in inputs:
            aug_path = utils_path._get_temp_filepath(
                path,
                prefix=prefix,
                suffix=suffix,
                add_uuid=add_uuid,
                add_timestamp=add_timestamp,
                ext=change_ext,
            )
            outputs.append(aug_path)

    # Output is a folder
    elif isinstance(output_path, str) and utils_path._check_is_dir(output_path):

        for path in inputs:
            aug_path = utils_path._get_augmented_path(
                path,
                prefix=prefix,
                suffix=suffix,
                add_uuid=add_uuid,
                add_timestamp=add_timestamp,
                change_ext=change_ext,
                folder=output_path,
            )
            outputs.append(aug_path)

    # Output is a file
    elif isinstance(output_path, str):
        if not utils_path._check_is_valid_output_filepath(output_path):
            raise ValueError("Invalid output_path.")

        aug_path = utils_path._get_augmented_path(
            output_path,
            prefix=prefix,
            suffix=suffix,
            add_uuid=add_uuid,
            add_timestamp=add_timestamp,
            change_ext=change_ext,
            folder=None,
        )
        outputs.append(aug_path)

    # Output is a list of files
    elif isinstance(output_path, list):
        if not utils_path._check_is_valid_filepath_list(output_path):
            raise ValueError("Invalid output_path.")

        if len(output_path) != len(inputs):
            raise ValueError("The number of output paths must match the number of input paths.")

        for path in output_path:
            aug_path = utils_path._get_augmented_path(
                path,
                prefix=prefix,
                suffix=suffix,
                add_uuid=add_uuid,
                add_timestamp=add_timestamp,
                change_ext=change_ext,
                folder=None,
            )
            outputs.append(aug_path)

    else:
        raise ValueError("Invalid output_path.")

    return outputs



"""
This module contains functions for augmenting images that are
suited to remote sensing imagery.
"""
# Standard library
import sys; sys.path.append("../../")
from typing import Optional

# External
import numpy as np

# Internal
from buteo.array.filters import filter_operation
from buteo.array.convolution_kernels import kernel_base


def spatial_label_smoothing(
    arr: np.array,
    radius: int = 2, *,
    classes: Optional[np.array] = None,
    method: Optional[str] = "half",
    variance: Optional[np.ndarray] = None,
    channel_last: bool = True,
) -> np.ndarray:
    """
    Smoothes the labels in a landcover classification by counting the weighted
    occurances of classes in a given radius.

    Parameters
    ----------
    arr : np.ndarray
        The array to smooth. Should only contain integer values and 1 channel.

    radius : int, optional
        The radius of the smoothing kernel, default: 2.

    classes : np.array, optional
        The classes in the classification. Numpy integer list of unique values in the array if None, default: None.

    method : str, optional
        Determines the flip protection method, default: 'half'.
        kernel: The class is protected by ensuring that the weight of the center class is always at least the sum of the surrounding weights.
        half: The class is protected by ensuring that it is always at least 50% + strength of the weight of the classes in the neighbourhood.
        max: The class is protected by ensuring that it is always at at least the max + strength of the strongest class in the neighbourhood.
        None: No flip protection.

    channel_last : bool, optional
        Whether the image is (channels, height, width) or (height, width, channels), default: True.
    
    Returns
    np.ndarray
        The smoothed array. One channel for each class in ascending order. Float32.
    """
    if classes is None:
        classes = np.unique(arr)

    if channel_last:
        classes_shape = np.array(classes).reshape(1, 1, -1)
    else:
        classes_shape = np.array(classes).reshape(-1, 1, 1)

    classes_hot = (arr == classes_shape).astype(np.uint8)

    if variance is not None:
        classes_arr = classes_hot * variance
    else:
        classes_arr = classes_hot

    kernel = kernel_base(
        radius=radius,
        circular=True,
        distance_weighted=True,
        normalised=False,
        hole=True if method is not None else False,
        method=3,
        sigma=2,
    )

    strength = np.float32(kernel.size / (kernel.size - 1.0))

    if method == "kernel":
        kernel[kernel.shape[0] // 2, kernel.shape[1] // 2] = kernel.sum() * strength

    feathered = filter_operation(classes_arr, 1, radius=radius, normalised=False, kernel=kernel, channel_last=channel_last)

    if method == "max":
        if channel_last:
            maxval = np.max(feathered, axis=2, keepdims=True)
            argmax = np.argmax(feathered, axis=2, keepdims=True)
        else:
            maxval = np.max(feathered, axis=0, keepdims=True)
            argmax = np.argmax(feathered, axis=0, keepdims=True)

        argmax_hot = np.argmax(classes_hot, axis=2, keepdims=True)

        weight = np.where(argmax == argmax_hot, feathered, maxval * strength)
        dst = np.where(classes_hot == 1, weight, feathered)
    
    elif method == "half":
        weight = kernel.sum() * strength 
        dst = np.where(classes_hot == 1, weight, feathered)

    elif method is None or method == "kernel":
        dst = feathered

    else:
        raise ValueError("Invalid method.")
    
    dst = dst / np.maximum(np.sum(dst, axis=2, keepdims=True), 1e-7)

    return dst



"""
### Extract the footprints or centroids of rasters. ###
"""

# Standard library
import sys; sys.path.append("../../")
from typing import Union, Optional, List
import os

# External
from osgeo import gdal

# Internal
from buteo.raster import core_raster
from buteo.utils import (
    utils_base,
    utils_bbox,
    utils_path,
    utils_io,
    utils_projection,
)



def raster_get_footprints(
    raster: Union[str, gdal.Dataset, List[Union[str, gdal.Dataset]]],
    latlng: bool = True,
    out_path: Optional[str] = None,
    *,
    overwrite: bool = True,
    prefix: str = "",
    suffix: str = "",
    add_uuid: bool = False,
    add_timestamp: bool = False,
    out_format: str = "gpkg",
) -> Union[str, List[str], gdal.Dataset]:
    """
    Gets the footprints of a raster or a list of rasters.

    Parameters
    ----------
    raster : Union[str, List, gdal.Dataset]
        The raster(s) to be shifted.

    latlng : bool, optional
        If True, the footprints are returned in lat/lon coordinates. If False, the footprints are returned in projected coordinates., default: True
        
    out_path : Optional[str], optional
        The path to the output raster. If None, the raster is created in memory., default: None

    overwrite : bool, optional
        If True, the output raster will be overwritten if it already exists., default: True

    prefix : str, optional
        The prefix to be added to the output raster name., default: ""

    suffix : str, optional
        The suffix to be added to the output raster name., default: ""

    add_uuid : bool, optional
        If True, a unique identifier will be added to the output raster name., default: False

    add_timestamp : bool, optional
        If True, a timestamp will be added to the output raster name., default: False
    
    out_format : str, optional
        The output format of the raster. If None, the format is inferred from the output path., default: "gpkg"

    Returns
    -------
    Union[str, List[str]]
        The path(s) to the shifted raster(s).
    """
    utils_base._type_check(raster, [str, gdal.Dataset, [str, gdal.Dataset]], "raster")
    utils_base._type_check(latlng, [bool], "latlng")
    utils_base._type_check(out_path, [str, [str], None], "out_path")
    utils_base._type_check(overwrite, [bool], "overwrite")
    utils_base._type_check(prefix, [str], "prefix")
    utils_base._type_check(suffix, [str], "suffix")
    utils_base._type_check(add_uuid, [bool], "add_uuid")
    utils_base._type_check(add_timestamp, [bool], "add_timestamp")

    input_is_list = isinstance(raster, list)

    input_list = utils_io._get_input_paths(raster, "raster")
    output_list = utils_io._get_output_paths(
        input_list,
        out_path,
        overwrite=overwrite,
        prefix=prefix,
        suffix=suffix,
        add_uuid=add_uuid,
        add_timestamp=add_timestamp,
        change_ext=out_format,
    )

    utils_path._delete_if_required_list(output_list, overwrite)

    footprints = []
    for idx, in_raster in enumerate(input_list):
        metadata = core_raster._get_basic_metadata_raster(in_raster)
        name = os.path.splitext(os.path.basename(metadata["name"]))[0]

        # Projections
        projection_osr = metadata["projection_osr"]
        projection_latlng = utils_projection._get_default_projection_osr()

        # Bounding boxes
        bbox = metadata["bbox"]

        if latlng:
            footprint_geom = utils_bbox._get_bounds_from_bbox(
                bbox,
                projection_osr,
                wkt=False,
            )
            footprint = utils_bbox._get_vector_from_geom(
                footprint_geom,
                projection_osr=projection_latlng,
                out_path=output_list[idx],
                name=name,
            )
        else:
            footprint_geom = utils_bbox._get_geom_from_bbox(
                bbox,
            )
            footprint = utils_bbox._get_vector_from_geom(
                footprint_geom,
                projection_osr=projection_osr,
                out_path=output_list[idx],
                name=name,
            )

        footprints.append(footprint)

    if input_is_list:
        return footprints

    return footprints[0]



"""
This module provides functions to ease the preprocessing of Sentinel 1 data
and finding the GPT tools.

TODO:
    - Improve documentation
"""

import xml.etree.ElementTree as ET
from datetime import datetime
from typing import Union, List, Tuple, Dict, Any
import numpy as np


# TODO: wkt to ogr geometry
def _s1_kml_to_bbox(path_to_kml: str) -> str:
    """ Internal. """
    root = ET.parse(path_to_kml).getroot()
    for elem in root.iter():
        if elem.tag == "coordinates":
            coords = elem.text
            break

    coords = coords.split(",")
    coords[0] = coords[-1] + " " + coords[0]
    del coords[-1]
    coords.append(coords[0])

    min_x = 180
    max_x = -180
    min_y = 90
    max_y = -90

    for coord in coords:
        intermediate = coord.split(" ")
        intermediate.reverse()

        intermediate[0] = float(intermediate[0])
        intermediate[1] = float(intermediate[1])

        if intermediate[0] < min_x:
            min_x = intermediate[0]
        elif intermediate[0] > max_x:
            max_x = intermediate[0]

        if intermediate[1] < min_y:
            min_y = intermediate[1]
        elif intermediate[1] > max_y:
            max_y = intermediate[1]

    footprint = f"POLYGON (({min_x} {min_y}, {min_x} {max_y}, {max_x} {max_y}, {max_x} {min_y}, {min_x} {min_y}))"

    return footprint


# TODO: get_metadata_from_zip
def s1_get_metadata(
    image_paths: List[str],
) -> List[Dict[str, Any]]:
    """
    Get metadata from Sentinel 1 images.

    Parameters
    ----------
    image_paths : List[str]
        List of paths to Sentinel 1 images.

    Returns
    -------
    List[Dict[str, Any]]
        List of metadata dictionaries.
    """
    images_obj = []

    for img in image_paths:
        kml = f"{img}/preview/map-overlay.kml"

        timestr = str(img.rsplit(".")[0].split("/")[-1].split("_")[5])
        timestamp = datetime.strptime(timestr, "%Y%m%dT%H%M%S").timestamp()

        meta = {
            "path": img,
            "timestamp": timestamp,
            "footprint_wkt": _s1_kml_to_bbox(kml),
        }

        images_obj.append(meta)

    return images_obj

def s1_to_db(
    arr: np.ndarray,
) -> np.ndarray:
    """
    Convert intensity to dB
    10 * log10(arr)
    """
    epsilon = np.finfo(arr.dtype).eps
    return 10.0 * np.log10(np.where(arr == 0.0, epsilon, arr))


def s1_to_intensity(
    arr: np.ndarray,
) -> np.ndarray:
    """
    Convert dB to intensity
    10^(arr/10)
    """
    return np.power(10, arr / 10.0)



"""
Various utility functions to work with the underlying systems.
"""
from .utils_base import *
from .utils_path import *
from .utils_aux import *
from .utils_gdal import *
from .utils_translate import *
from .utils_projection import *
from .utils_bbox import *
from .utils_io import *



"""
This module provides varoius indices for Sentinel 2 imagery.

TODO:
    Convert to numpy functions.
    Handle resampling
"""
# Standard library
import sys; sys.path.append("../../")
from typing import List

# External
import numpy as np


def _all_arrays_are_same_size(arrays: List[np.ndarray]) -> bool:
    """ Check if all arrays are the same size. """
    return all([arrays[0].shape == array.shape for array in arrays])

def s2_index_chlre(b_05: np.ndarray, b_08: np.ndarray):
    """ Red Edge Chlorophyll Index
        Calc: B05 / B08
    """
    assert _all_arrays_are_same_size([b_05, b_08]), "All arrays must be the same size."

    result = np.zeros_like(b_08, dtype=np.float32)
    np.divide(b_05, b_08, out=result, where=b_08 != 0)

    return result

def s2_index_rendvi(b_08: np.ndarray, b_06: np.ndarray):
    """ Red Edge NDVI
        Calc: (B08 - B06) / (B08 + B06)
    """
    assert _all_arrays_are_same_size([b_08, b_06]), "All arrays must be the same size."

    result = np.zeros_like(b_08, dtype=np.float32)

    add = b_08 + b_06
    sub = b_08 - b_06
    np.divide(sub, add, out=result, where=add != 0)

    return result

def s2_index_s2_rep(b_07: np.ndarray, b_04: np.ndarray, b_05: np.ndarray, b_06: np.ndarray):
    """ Sentinel 2 Red Edge Position
        Calc: 705 + 35 * ((((B07 + B04) / 2) - B05) / (B06 - B05))
    """
    assert _all_arrays_are_same_size([b_07, b_04, b_05, b_06]), "All arrays must be the same size."

    result = np.zeros_like(b_07, dtype=np.float32)

    sub = b_06 - b_05
    np.divide(705 + 35 * ((((b_07 + b_04) / 2) - b_05) / sub), 1, out=result, where=sub != 0)

    return result

def s2_index_ireci(b_07: np.ndarray, b_04: np.ndarray, b_05: np.ndarray, b_06: np.ndarray):
    """ Red Edge Chlorophyll Index
        Calc: (B07 - B04) * B06 / B05
    """
    assert _all_arrays_are_same_size([b_07, b_04, b_05, b_06]), "All arrays must be the same size."

    result = np.zeros_like(b_07, dtype=np.float32)
    np.divide((b_07 - b_04) * b_06, b_05, out=result, where=b_05 != 0)

    return result

def s2_index_mcari(b_05: np.ndarray, b_04: np.ndarray, b_03: np.ndarray):
    """ Modified Chlorophyll Absorption in Reflectance Index
        Calc: (B05 - B04) - 0.2 * (B05 - B03) * (B05 / B04)
    """
    assert _all_arrays_are_same_size([b_05, b_04, b_03]), "All arrays must be the same size."

    return (b_05 - b_04) - 0.2 * (b_05 - b_03) * (b_05 / b_04)

def s2_index_arvi(b_08: np.ndarray, b_04: np.ndarray, b_02: np.ndarray):
    """ Atmospherically Resistant Vegetation Index
        Calc: (B08 - b) / (B08 + b), b = 2 * B04 - B02
    """
    assert _all_arrays_are_same_size([b_08, b_04, b_02]), "All arrays must be the same size."

    b_calc = 2 * b_04 - b_02
    sub = b_08 - b_calc
    add = b_08 + b_calc

    result = np.zeros_like(b_08, dtype=np.float32)
    np.divide(sub, add, out=result, where=add != 0)

    return result

def s2_index_savi(b_08: np.ndarray, b_04: np.ndarray):
    """ Soil Adjusted Vegetation Index
        Calc: ((B08 - B04) / ((B08 + B04) + 0.428)) * 1.856
    """
    assert _all_arrays_are_same_size([b_08, b_04]), "All arrays must be the same size."

    sub = b_08 - b_04
    add = b_08 + b_04 + 0.428

    result = np.zeros_like(b_08, dtype=np.float32)
    np.divide(sub, add, out=result, where=add != 0)

    return result * 1.856

def s2_index_msavi2(b_08: np.ndarray, b_04: np.ndarray):
    """ Modified Soil Adjusted Vegetation Index v2
        Calc: (2 * B08 + 1 - sqrt((2 * B08 + 1) ** 2) - 8 * (B08 - B04)) / 2
    """
    assert _all_arrays_are_same_size([b_08, b_04]), "All arrays must be the same size."

    return (2 * b_08 + 1 - np.sqrt((2 * b_08 + 1) ** 2) - 8 * (b_08 - b_04)) / 2

def s2_index_gndvi(b_08: np.ndarray, b_03: np.ndarray):
    """ Green Normalized Difference Vegetation Index
        Calc: (B08 - B03) / (B08 + B03)
    """
    assert _all_arrays_are_same_size([b_08, b_03]), "All arrays must be the same size."

    result = np.zeros_like(b_08, dtype=np.float32)

    add = b_08 + b_03
    sub = b_08 - b_03
    np.divide(sub, add, out=result, where=add != 0)

    return result

def s2_index_ndvi(b_08: np.ndarray, b_04: np.ndarray):
    """ Normalized Difference Vegetation Index
        Calc: (B08 - B04) / (B08 + B04)
    """
    assert _all_arrays_are_same_size([b_08, b_04]), "All arrays must be the same size."

    result = np.zeros_like(b_08, dtype=np.float32)

    add = b_08 + b_04
    sub = b_08 - b_04
    np.divide(sub, add, out=result, where=add != 0)

    return result

def s2_index_moist(b_8A: np.ndarray, b_11: np.ndarray):
    """ Soil Moisture Index
        Calc: (B8A - B11) / (B8A + B11)
    """
    assert _all_arrays_are_same_size([b_8A, b_11]), "All arrays must be the same size."

    result = np.zeros_like(b_8A, dtype=np.float32)

    add = b_8A + b_11
    sub = b_8A - b_11
    np.divide(sub, add, out=result, where=add != 0)

    return result

def s2_index_ndwi(b_08: np.ndarray, b_11: np.ndarray):
    """ Normalized Difference Water Index
        Calc: (B08 - B11) / (B08 + B11)
    """
    assert _all_arrays_are_same_size([b_08, b_11]), "All arrays must be the same size."

    result = np.zeros_like(b_08, dtype=np.float32)

    add = b_08 + b_11
    sub = b_08 - b_11
    np.divide(sub, add, out=result, where=add != 0)

    return result

def s2_index_ndwi_v2(b_03: np.ndarray, b_08: np.ndarray):
    """ Normalized Difference Water Index v2
        Calc: (B03 - B08) / (B03 + B08)
    """
    assert _all_arrays_are_same_size([b_03, b_08]), "All arrays must be the same size."

    result = np.zeros_like(b_03, dtype=np.float32)

    add = b_03 + b_08
    sub = b_03 - b_08
    np.divide(sub, add, out=result, where=add != 0)

    return result

def s2_index_nbr(b_08: np.ndarray, b_12: np.ndarray):
    """ Normalized Burn Ratio
        Calc: (B08 - B12) / (B08 + B12)
    """
    assert _all_arrays_are_same_size([b_08, b_12]), "All arrays must be the same size."

    result = np.zeros_like(b_08, dtype=np.float32)

    add = b_08 + b_12
    sub = b_08 - b_12
    np.divide(sub, add, out=result, where=add != 0)

    return result

def s2_index_nvei(b_02: np.ndarray, b_04: np.ndarray):
    """ Non-elimination vegetation index
        Calc: (B02 - B04) / (B08 + B04)
    """
    assert _all_arrays_are_same_size([b_02, b_04]), "All arrays must be the same size."

    result = np.zeros_like(b_02, dtype=np.float32)

    add = b_02 + b_04
    sub = b_02 - b_04
    np.divide(sub, add, out=result, where=add != 0)

    return result

def s2_index_nbai(b_02: np.ndarray, b_08: np.ndarray, b_12: np.ndarray): # Built-up area index
    """ Built-up area index
        Calc: (B12 - d) / (B12 + d), d = B08 / B02
    """
    assert _all_arrays_are_same_size([b_02, b_08, b_12]), "All arrays must be the same size."

    intermediate = np.zeros_like(b_02, dtype=np.float32)
    result = np.zeros_like(b_02, dtype=np.float32)

    np.divide(b_08, b_02, out=intermediate, where=b_02 != 0)

    add = b_12 + intermediate
    sub = b_12 - intermediate
    np.divide(sub, add, out=result, where=add != 0)

    return result

def s2_index_brba(b_03: np.ndarray, b_08: np.ndarray):
    """ Band ratio for built-up areas
        Calc: (B03 / B08)
    """
    assert _all_arrays_are_same_size([b_03, b_08]), "All arrays must be the same size."

    result = np.zeros_like(b_03, dtype=np.float32)
    np.divide(b_03, b_08, out=result, where=b_08 != 0)

    return result

def s2_index_ndbi(b_11: np.ndarray, b_08: np.ndarray):
    """ Normalised difference built-up index
        Calc: (B11 - B08) / (B11 + B08)
    """
    assert _all_arrays_are_same_size([b_11, b_08]), "All arrays must be the same size."

    result = np.zeros_like(b_11, dtype=np.float32)

    add = b_11 + b_08
    sub = b_11 - b_08
    np.divide(sub, add, out=result, where=add != 0)

    return result

def s2_index_blfei(b_03: np.ndarray, b_04: np.ndarray, b_11: np.ndarray, b_12: np.ndarray):
    """ Built-up features extraction
        Calc: (bix - B11) / (bix + B11), bix = (B03 + B04 + B12) / 3
    """
    assert _all_arrays_are_same_size([b_03, b_04, b_11, b_12]), "All arrays must be the same size."

    result = np.zeros_like(b_11, dtype=np.float32)

    bix_calc = (b_03 + b_04 + b_12) / 3

    add = bix_calc + b_11
    sub = bix_calc - b_11
    np.divide(sub, add, out=result, where=add != 0)

    return result

def s2_index_ibi(b_03: np.ndarray, b_04: np.ndarray, b_08: np.ndarray, b_11: np.ndarray):
    """ Built-up features extraction
        Calc: (ndbi - ((savi + ndwi2) / 2)) / (ndbi + ((savi + ndwi2) / 2))
            savi = ((B08 - B04) / ((B08 + B04) + 0.428)) * 1.856
            ndwi2 = (B03 - B08) / (B03 + B08)
            ndbi = (B11 - B08) / (B11 + B08
    """
    assert _all_arrays_are_same_size([b_03, b_04, b_08, b_11]), "All arrays must be the same size."

    savi_calc = s2_index_savi(b_08, b_04)
    ndwi2_calc = s2_index_ndwi(b_03, b_08)
    ndbi_calc = s2_index_ndbi(b_11, b_08)

    result = np.zeros_like(b_11, dtype=np.float32)

    add = ndbi_calc + (savi_calc + ndwi2_calc) / 2
    sub = ndbi_calc - (savi_calc + ndwi2_calc) / 2

    np.divide(sub, add, out=result, where=add != 0)

    return result



"""
This module provides a set of functions to normalise data for machine learning.
"""

# Standard library
from typing import Union, Optional, Tuple

# External
import numpy as np


def scaler_minmax(
    arr: np.array,
    stat_dict: Optional[dict] = None,
) -> Tuple[np.array, dict]:
    """
    Normalize an input numpy array using the min-max scaling method.

    Parameters
    ----------
    arr : np.ndarray
        The input numpy array to be normalized.

    stat_dict : Optional[Dict[str, float]], default: None
        A dictionary containing the minimum and maximum values of the input array.
        If not provided, the function will compute the minimum and maximum values from the input array.

    Returns
    -------
    Tuple[np.ndarray, Dict[str, float]]
        A tuple containing the normalized numpy array (with dtype float32) and a dictionary containing the minimum
        and maximum values used for scaling. The minimum and maximum values are stored in the dictionary using the
        keys "min" and "max", respectively.

    Raises
    ------
    AssertionError
        If the input array is not a numpy array, or if the stat_dict is not a dictionary or None.
    """
    assert isinstance(arr, np.ndarray), "Input must be a numpy array."
    assert isinstance(stat_dict, dict) or stat_dict is None, "stat_dict must be a dictionary or None"

    if stat_dict is not None:
        arr_min = stat_dict["min"]
        arr_max = stat_dict["max"]
    else:
        arr_min = arr.min()
        arr_max = arr.max()

    stat_dict = {
        "min": arr.min(),
        "max": arr.max(),
    }

    num = np.subtract(arr, arr_min)
    den = np.subtract(arr_max, arr_min)

    result = np.zeros_like(arr, dtype="float32")
    np.divide(num, den, out=result, where=den != 0)

    return result, stat_dict


def scaler_standardise(
    arr: np.array,
    stat_dict: Optional[dict] = None,
) -> Tuple[np.array, dict]:
    """
    Standardize an an input numpy array.

    Parameters
    ----------
    arr : np.ndarray
        The input numpy array to be normalized.

    stat_dict : Optional[Dict[str, float]], default: None
        A dictionary containing the mean and standard deviation values of the input array.
        If not provided, the function will compute the mean and standard deviation from the input array.

    Returns
    -------
    Tuple[np.ndarray, Dict[str, float]]
        A tuple containing the normalized numpy array (with dtype float32) and a dictionary containing the mean
        and standard deviation values used for scaling. The mean and standard deviation values are stored in the
        dictionary using the keys "mean" and "std", respectively.

    Raises
    ------
    AssertionError
        If the input array is not a numpy array, or if the stat_dict is not a dictionary or None.
    """
    assert isinstance(arr, np.ndarray), "Input must be a numpy array."
    assert isinstance(stat_dict, dict) or stat_dict is None, "stat_dict must be a dictionary or None"

    if stat_dict is not None:
        arr_mean = stat_dict["mean"]
        arr_std = stat_dict["std"]
    else:
        arr_mean = np.nanmean(arr)
        arr_std = np.nanstd(arr)

    stat_dict = {
        "mean": arr_mean,
        "std": arr_std,
    }

    result = np.zeros_like(arr, dtype="float32")

    num = np.subtract(arr, arr_mean)
    np.divide(num, arr_std, out=result, where=arr_std != 0)

    return result, stat_dict


def scaler_standardise_mad(
    arr: np.array,
    stat_dict: Optional[dict] = None,
) -> Tuple[np.array, dict]:
    """
    Normalize an input numpy array using the median absolute deviation (MAD) scaling method.

    Parameters
    ----------
    arr : np.ndarray
        The input numpy array to be normalized.

    stat_dict : Optional[Dict[str, float]], default: None
        A dictionary containing the median, absolute deviation and median absolute deviation of the input array.
        If not provided, the function will compute these statistics from the input array.

    Returns
    -------
    Tuple[np.ndarray, Dict[str, float]]
        A tuple containing the normalized numpy array (with dtype float32) and a dictionary containing the median,
        absolute deviation, and median absolute deviation used for scaling. These statistics are stored in the
        dictionary using the keys "median", "absdev", and "madstd", respectively.

    Raises
    ------
    AssertionError
        If the input array is not a numpy array, or if the stat_dict is not a dictionary or None.
    """
    assert isinstance(arr, np.ndarray), "Input must be a numpy array."
    assert isinstance(stat_dict, dict) or stat_dict is None, "stat_dict must be a dictionary or None"

    if stat_dict is not None:
        median = stat_dict["median"]
        absdev = stat_dict["absdev"]
        madstd = stat_dict["madstd"]
    else:
        median = np.nanmedian(arr)
        absdev = np.abs(np.subtract(arr, median))
        madstd = np.nanmedian(absdev) * 1.4826

    stat_dict = {
        "median": median,
        "absdev": absdev,
        "madstd": madstd,
    }

    result = np.zeros_like(arr, dtype="float32")

    sub = np.subtract(arr, median)
    np.divide(sub, madstd, out=result, where=madstd != 0)

    return result, stat_dict


def scaler_iqr(
    arr: np.array,
    q1=0.25,
    q3=0.75,
    stat_dict: Optional[dict] = None,
) -> Tuple[np.array, dict]:
    """ 
    Normalize an input numpy array using the interquartile range (IQR) method.

    Parameters
    ----------
    arr : np.ndarray
        The input numpy array to be normalized.

    q1 : float, default: 0.25
        The lower quartile to use in the IQR calculation.

    q3 : float, default: 0.75
        The upper quartile to use in the IQR calculation.

    stat_dict : Optional[Dict[str, float]], default: None
        A dictionary containing the lower, median, and upper quartiles of the input array.
        If not provided, the function will compute the quartiles from the input array.

    Returns
    -------
    Tuple[np.ndarray, Dict[str, float]]
        A tuple containing the normalized numpy array (with dtype float32) and a dictionary containing the lower,
        median, and upper quartiles used for scaling. The quartile values are stored in the dictionary using the keys
        "q1", "median", and "q3", respectively.

    Raises
    ------
    AssertionError
        If the input array is not a numpy array, q1 is not between 0.0 and 0.5, q3 is not between 0.5 and 1.0, or if
        q1 is greater than or equal to q3, or if stat_dict is not a dictionary or None.
    """
    assert isinstance(arr, np.ndarray), "Input must be a numpy array."
    assert 0.0 <= q1 < 0.5, "q1 must be less than 0.5 and above or equal to 0.0"
    assert 0.5 < q3 <= 1.0, "q3 must be greater than 0.5 and below or equal to 1.0"
    assert q1 < q3, "q1 must be less than q3"
    assert isinstance(stat_dict, dict) or stat_dict is None, "stat_dict must be a dictionary or None"

    if stat_dict is not None:
        q1 = stat_dict["q1"]
        median = stat_dict["median"]
        q3 = stat_dict["q3"]
    else:
        q1, median, q3 = np.nanquantile(arr, [q1, 0.5, q3])

    stat_dict = {
        "q1": q1,
        "median": median,
        "q3": q3,
    }

    num = np.subtract(arr, median)
    den = np.subtract(q3, q1)

    result = np.zeros_like(arr, dtype="float32")
    np.divide(num, den, out=result, where=den != 0)

    return result, stat_dict


def scaler_to_range(
    arr: np.ndarray,
    min_val: Union[float, int] = 0.0,
    max_val: Union[float, int] = 1.0,
    stat_dict: Optional[dict] = None,
) -> Tuple[np.array, dict]:
    """
    Normalizes an input numpy array to a given range.

    Parameters
    ----------
    arr : np.ndarray
        The input numpy array to be normalized.

    min_val : Union[float, int], optional
        The minimum value to scale the array to, default: 0.0.

    max_val : Union[float, int], optional
        The maximum value to scale the array to, default: 1.0.

    stat_dict : Optional[Dict[str, float]], optional
        A dictionary containing the minimum and maximum values of the input array.
        If not provided, the function will compute the minimum and maximum values from the input array.
        default: None.

    Returns
    -------
    Tuple[np.array, Dict[str, float]]
        A tuple containing the normalized numpy array (with dtype float32) and a dictionary containing the minimum
        and maximum values used for scaling. The minimum and maximum values are stored in the dictionary using the
        keys "min" and "max", respectively.

    Raises
    ------
    AssertionError
        If the input array is not a numpy array or if the stat_dict is not a dictionary or None.

    Notes
    -----
    This function uses the formula (arr - arr_min) / (arr_max - arr_min) to normalize the input array to a range
    of [min_val, max_val]. The minimum and maximum values of the input array can be passed through stat_dict,
    otherwise, they will be computed from the array.

    Examples
    --------
    ```python
    >>> arr = np.array([1, 2, 3, 4, 5])
    >>> scaler_to_range(arr, 0, 10)
    (array([0., 2., 4., 6., 10.], dtype=float32), {'min': 1, 'max': 5})
    ```
    """
    assert isinstance(arr, np.ndarray), "Input must be a numpy array."
    assert isinstance(stat_dict, dict) or stat_dict is None, "stat_dict must be a dictionary or None"

    if stat_dict is not None:
        arr_min = stat_dict["min"]
        arr_max = stat_dict["max"]
    else:
        arr_min = np.min(arr)
        arr_max = np.max(arr)

    stat_dict = {
        "min": arr_min,
        "max": arr_max,
    }

    num = np.subtract(arr, arr_min)
    den = np.subtract(arr_max, arr_min)

    result = np.zeros_like(arr, dtype="float32")
    np.divide(num, den, out=result, where=den != 0)

    result = np.multiply(result, max_val - min_val) + min_val

    return result, stat_dict


def scaler_truncate(
    arr: np.ndarray,
    trunc_min: Union[float, int] = None,
    trunc_max: Union[float, int] = None,
    target_min: Union[float, int] = 0.0,
    target_max: Union[float, int] = 1.0,
    stat_dict: Optional[dict] = None,
) -> Tuple[np.array, dict]:
    """
    Truncate an input numpy array within the given range, and then normalise it to the target range.

    Parameters
    ----------
    arr : np.ndarray
        The input numpy array to be truncated and normalized.

    trunc_min : float/int, optional
        The minimum value to truncate to. If not provided, truncation will not be applied to the lower end.

    trunc_max : float/int, optional
        The maximum value to truncate to. If not provided, truncation will not be applied to the upper end.

    target_min : float/int, optional
        The minimum value of the target range to normalize to. Default: 0.0.

    target_max : float/int, optional
        The maximum value of the target range to normalize to. Default: 1.0.

    stat_dict : Dict[str, float], optional
        A dictionary containing the minimum and maximum values of the input array. If not provided, the function
        will compute the minimum and maximum values from the truncated array.

    Returns
    -------
    Tuple[np.array, Dict[str, float]]
        A tuple containing the normalized numpy array (with dtype float32) and a dictionary containing the minimum
        and maximum values used for scaling. The minimum and maximum values are stored in the dictionary using the
        keys "arr_min" and "arr_max", respectively.

    Raises
    ------
    AssertionError
        If the input array is not a numpy array, or if the trunc_min is not less than trunc_max, or if the
        target_min is not less than target_max, or if the stat_dict is not a dictionary or None.
    """
    assert isinstance(arr, np.ndarray), "Input must be a numpy array."
    assert isinstance(stat_dict, dict) or stat_dict is None, "stat_dict must be a dictionary or None"
    assert trunc_min < trunc_max, "trunc_min must be less than trunc_max"
    assert target_min < target_max, "target_min must be less than target_max"

    if stat_dict is not None:
        trunc_min = stat_dict["trunc_min"]
        trunc_max = stat_dict["trunc_max"]
        target_min = stat_dict["target_min"]
        target_max = stat_dict["target_max"]
    else:
        stat_dict = {
            "trunc_min": trunc_min,
            "trunc_max": trunc_max,
            "target_min": target_min,
            "target_max": target_max,
        }

    truncated = arr
    if trunc_min is not None:
        truncated = np.where(arr > trunc_max, trunc_max, arr)

    if trunc_max is not None:
        truncated = np.where(truncated < trunc_min, trunc_min, truncated)

    if "arr_min" not in stat_dict:
        stat_dict["arr_min"] = np.min(truncated)
    if "arr_max" not in stat_dict:
        stat_dict["arr_max"] = np.max(truncated)

    truncated = truncated.astype(np.float32, copy=False)

    arr_min = stat_dict["arr_min"]
    arr_max = stat_dict["arr_max"]

    num = np.subtract(truncated, arr_min)
    den = np.subtract(arr_max, arr_min)

    result = np.zeros_like(arr, dtype="float32")
    np.divide(num, den, out=result, where=den != 0)

    result = np.multiply(result, target_max - target_min) + target_min

    return result, stat_dict



# Standard library
import sys; sys.path.append("../../")
from typing import Union, List, Dict, Any
import os

# External
from osgeo import ogr, gdal, osr

# Internal
from buteo.utils import (
    utils_io,
    utils_base,
    utils_gdal,
    utils_bbox,
    utils_path,
)
from buteo.vector.core_vector import _vector_open



def _vector_to_metadata(
    vector: Union[str, ogr.DataSource, gdal.Dataset],
) -> Dict[str, Any]:
    """ Internal. """
    assert isinstance(vector, (str, ogr.DataSource)), "vector must be a path or a DataSource"

    datasource = _vector_open(vector)

    vector_driver = datasource.GetDriver()

    path = datasource.GetDescription()
    basename = os.path.basename(path)
    split_path = os.path.split(basename)
    name = split_path[0]
    ext = split_path[1]

    driver = vector_driver.GetName()

    in_memory = utils_gdal._check_is_dataset_in_memory(datasource)

    layer_count = datasource.GetLayerCount()
    layers = []

    vector_bbox = None
    for layer_index in range(layer_count):
        layer = datasource.GetLayerByIndex(layer_index)

        x_min, x_max, y_min, y_max = layer.GetExtent()
        layer_bbox = [x_min, x_max, y_min, y_max]
        vector_bbox = layer_bbox

        layer_name = layer.GetName()

        column_fid = layer.GetFIDColumn()
        column_geom = layer.GetGeometryColumn()

        if column_geom == "":
            column_geom = "geom"

        feature_count = layer.GetFeatureCount()

        projection_osr = layer.GetSpatialRef()
        projection_wkt = layer.GetSpatialRef().ExportToWkt()

        if layer_index > 0:
            v_x_min, v_x_max, v_y_min, v_y_max = vector_bbox
            if x_min < v_x_min:
                vector_bbox[0] = x_min
            if x_max > v_x_max:
                vector_bbox[1] = x_max
            if y_min < v_y_min:
                vector_bbox[2] = y_min
            if y_max > v_y_max:
                vector_bbox[3] = y_max

        layer_defn = layer.GetLayerDefn()

        geom_type_ogr = layer_defn.GetGeomType()
        geom_type = ogr.GeometryTypeToName(layer_defn.GetGeomType())

        field_count = layer_defn.GetFieldCount()
        field_names = []
        field_types = []
        field_types_ogr = []

        for field_index in range(field_count):
            field_defn = layer_defn.GetFieldDefn(field_index)
            field_names.append(field_defn.GetName())
            field_type = field_defn.GetType()
            field_types_ogr.append(field_type)
            field_types.append(field_defn.GetFieldTypeName(field_type))

        layer_dict = {
            "layer_name": layer_name,
            "x_min": x_min,
            "x_max": x_max,
            "y_min": y_min,
            "y_max": y_max,
            "column_fid": column_fid,
            "column_geom": column_geom,
            "feature_count": feature_count,
            "projection_wkt": projection_wkt,
            "projection_osr": projection_osr,
            "geom_type": geom_type,
            "geom_type_ogr": geom_type_ogr,
            "field_count": field_count,
            "field_names": field_names,
            "field_types": field_types,
            "field_types_ogr": field_types_ogr,
            "is_raster": False,
            "is_vector": True,
            "bbox": layer_bbox,
            "extent": layer_bbox,
        }

        layer_bboxes = utils_bbox._additional_bboxes(layer_bbox, projection_osr)

        for key, value in layer_bboxes.items():
            layer_dict[key] = value


        ## MOVE TO A SINGLE FUNCTION
        def get_bbox_as_vector_layer():
            return utils_bbox._get_vector_from_bbox(layer_bbox, projection_osr) # pylint: disable=cell-var-from-loop


        def get_bbox_as_vector_latlng_layer():
            projection_osr_latlng = osr.SpatialReference()
            projection_osr_latlng.ImportFromWkt('GEOGCS["WGS 84",DATUM["WGS_1984",SPHEROID["WGS 84",6378137,298.257223563]],PRIMEM["Greenwich",0],UNIT["degree",0.0174532925199433,AUTHORITY["EPSG","9122"]],AUTHORITY["EPSG","4326"]]')
            # projection_osr_latlng.ImportFromEPSG(4326)

            return utils_bbox._get_vector_from_bbox(layer_dict["bbox_latlng"], projection_osr_latlng)  # pylint: disable=cell-var-from-loop


        layer_dict["get_bbox_vector"] = get_bbox_as_vector_layer
        layer_dict["get_bbox_vector_latlng"] = get_bbox_as_vector_latlng_layer

        layers.append(layer_dict)

    metadata = {
        "path": path,
        "basename": basename,
        "name": name,
        "ext": ext,
        "in_memory": in_memory,
        "projection_wkt": projection_wkt,
        "projection_osr": projection_osr,
        "driver": driver,
        "x_min": vector_bbox[0],
        "x_max": vector_bbox[1],
        "y_min": vector_bbox[2],
        "y_max": vector_bbox[3],
        "is_vector": True,
        "is_raster": False,
        "layer_count": layer_count,
        "layers": layers,
        "extent": vector_bbox,
        "bbox": vector_bbox,
    }

    vector_bboxes = utils_bbox._additional_bboxes(vector_bbox, projection_osr)

    for key, value in vector_bboxes.items():
        metadata[key] = value


    def get_bbox_as_vector():
        return utils_bbox._get_vector_from_bbox(vector_bbox, projection_osr) # pylint: disable=cell-var-from-loop


    def get_bbox_as_vector_latlng():
        projection_osr_latlng = osr.SpatialReference()
        projection_osr_latlng.ImportFromWkt('GEOGCS["WGS 84",DATUM["WGS_1984",SPHEROID["WGS 84",6378137,298.257223563]],PRIMEM["Greenwich",0],UNIT["degree",0.0174532925199433,AUTHORITY["EPSG","9122"]],AUTHORITY["EPSG","4326"]]')
        # projection_osr_latlng.ImportFromEPSG(4326)

        return utils_bbox._get_vector_from_bbox(metadata["bbox_latlng"], projection_osr_latlng)  # pylint: disable=cell-var-from-loop


    metadata["get_bbox_vector"] = get_bbox_as_vector
    metadata["get_bbox_vector_latlng"] = get_bbox_as_vector_latlng

    return metadata


def vector_to_metadata(
    vector: Union[ogr.DataSource, str, List[Union[ogr.DataSource, str]]],
    allow_lists: bool = True,
) -> Union[Dict[str, Any], List[Dict[str, Any]]]:
    """
    Creates a dictionary with metadata about the vector layer.

    Parameters
    ----------
    vector : Union[ogr.DataSource, str, List[Union[ogr.DataSource, str]]]
        A vector layer(s) or path(s) to a vector file.
    
    allow_lists : bool, optional
        If True, vector can be a list of vector layers or paths. If False, vector must be a single vector layer or path. default: True

    Returns
    -------
    Union[Dict[str, Any], List[Dict[str, Any]]]
    """
    utils_base._type_check(vector, [str, ogr.DataSource, [str, ogr.DataSource]], "vector")
    utils_base._type_check(allow_lists, [bool], "allow_lists")

    if isinstance(vector, list) and not allow_lists:
        raise ValueError("The vector parameter cannot be a list when allow_lists is False.")

    vector_list = utils_base._get_variable_as_list(vector)

    if not utils_gdal._check_is_vector_list(vector_list):
        raise ValueError("The vector parameter must be a list of vector layers.")

    output = []

    for in_vector in vector_list:
        output.append(_vector_to_metadata(in_vector))

    if isinstance(vector, list):
        return output

    return output[0]
